#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **hint logic class hierarchy** (i.e., dataclasses encapsulating
all low-level Python code snippets and associated metadata required to
dynamically generate high-level Python code snippets fully type-checking various
kinds of type hints uniquely identified by common signs).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from abc import (
    ABCMeta,
    abstractmethod,
)
from beartype.typing import (
    TYPE_CHECKING,
    Callable,
)
from beartype._check.code.snip.codesnipcls import PITH_INDEX_TO_VAR_NAME
from beartype._check.metadata.hint.hintsmeta import HintsMeta
from beartype._check.error.errcause import ViolationCause
from beartype._check.metadata.hint.hintsane import HintSane
from beartype._conf.confenum import BeartypeStrategy
from beartype._data.typing.datatyping import (
    CallableStrFormat,
    EnumeratorItem,
    Enumerator,
)
from beartype._data.code.pep.datacodepep484585 import (
    CODE_PEP484585_QUASIITERABLE_format,
    CODE_PEP484585_REITERABLE_OR_SEQUENCE_format,
    CODE_PEP484585_REITERABLE_PITH_CHILD_EXPR_format,
    CODE_PEP484585_SEQUENCE_PITH_CHILD_EXPR_format,
)
from collections.abc import (
    Collection as CollectionABC,
    Sequence as SequenceABC,
)

# ....................{ PRIVATE ~ hints                    }....................
_GetCauseEnumeratorItem = Callable[[ViolationCause], EnumeratorItem]
'''
PEP-compliant type hint matching an **enumerator item violation cause getter**
(i.e., callable accepting a :class:`.ViolationCause` object and returning a
2-tuple ``(item_index, item)`` describing an arbitrary item efficiently accessed
from the container encapsulated by this violation cause.

This hint matches callables with signatures resembling:

.. code-block:: python

   def _get_cause_enumerator_item(cause: ViolationCause) -> EnumeratorItem:

Callables matched by this hint return 2-tuples of the standard form
``(item_index, item)`` returned by the :func:`enumerate` builtin, where:

* ``item_index`` is the 0-based index of an arbitrary item efficiently accessed
  from this container.
* ``item`` is that item.
'''

# ....................{ SUPERCLASSES                       }....................
class HintLogicABC(object, metaclass=ABCMeta):
    '''
    Abstract base class (ABC) of all **hint logic** (i.e., dataclasses
    encapsulating all low-level Python code snippets and associated metadata
    required to dynamically generate high-level Python code snippets fully
    type-checking some kind of type hint uniquely identified by a common sign).

    Caveats
    -------
    **Python code snippets should not contain ternary conditionals.** For
    unknown reasons suggesting a critical defect in the current implementation
    of Python 3.8's assignment expressions, snippets containing one or more
    ternary conditionals raise :exc:`UnboundLocalError` exceptions resembling:

        UnboundLocalError: local variable '__beartype_pith_1' referenced before
        assignment

    In particular, the initial draft of these snippets guarded against empty
    sequences with a seemingly reasonable ternary conditional:

    .. code-block:: python

       CODE_PEP484585_SEQUENCE = \'\'\'(
       {indent_curr}    isinstance({pith_curr_assign_expr}, {hint_curr_expr}) and
       {indent_curr}    {hint_child_placeholder} if {pith_curr_var_name} else True
       {indent_curr})\'\'\'

    That should behave as expected, but doesn't, presumably due to obscure
    scoping rules and a non-intuitive implementation of ternary conditionals in
    CPython. Ergo, the current version of this snippet guards against empty
    sequences with disjunctions and conjunctions (i.e., ``or`` and ``and``
    operators) instead. Happily, the current version is more efficient than the
    equivalent approach based on ternary conditional (albeit less intuitive).

    Attributes
    ----------
    is_var_random_int_needed : bool
        :data:`True` only if the Python code snippet dynamically generated by
        calling the :attr:`code_format` method requires a pseudo-random integer
        by accessing the local variable named
        :data:`beartype._data.code.datacodename.VAR_NAME_RANDOM_INT`. If :data:`True`,
        the body of the current wrapper function will be prefixed by a Python
        statement assigning such an integer to that local variable.
    _get_cause_enumerator_item : _GetCauseEnumeratorItem
        **Enumerator item violation cause getter** (i.e., callable accepting a
        :class:`.ViolationCause` object and returning a 2-tuple ``(item_index,
        item)`` describing an arbitrary item efficiently accessed from the
        container encapsulated by this violation cause.
    '''

    # ..................{ CLASS VARIABLES                    }..................
    # Slot all instance variables defined on this object to minimize the time
    # complexity of both reading and writing variables across frequently called
    # cache dunder methods. Slotting has been shown to reduce read and write
    # costs by approximately ~10%, which is non-trivial.
    __slots__ = (
        '_get_cause_enumerator_item',
        'is_var_random_int_needed',
    )

    # Squelch false negatives from mypy. This is absurd. This is mypy. See:
    #     https://github.com/python/mypy/issues/5941
    if TYPE_CHECKING:
        _get_cause_enumerator_item: _GetCauseEnumeratorItem
        is_var_random_int_needed : bool

    # ..................{ INITIALIZERS                       }..................
    def __init__(
        self,

        # Mandatory parameters.
        get_cause_enumerator_item: _GetCauseEnumeratorItem,

        # Optional parameters.
        is_var_random_int_needed: bool = False,
    ) -> None:
        '''
        Initialize this hint logic.

        Parameters
        ----------
        See the class docstring for further details.
        '''
        assert isinstance(is_var_random_int_needed, bool), (
            f'{repr(is_var_random_int_needed)} not boolean.')
        assert callable(get_cause_enumerator_item), (
            f'{repr(get_cause_enumerator_item)} uncallable.')

        # Classify all passed parameters.
        self._get_cause_enumerator_item = get_cause_enumerator_item
        self.is_var_random_int_needed = is_var_random_int_needed

    # ..................{ ITERATORS                          }..................
    def enumerate_cause_items(self, cause: ViolationCause) -> Enumerator:
        '''
        Arbitrary iterator satisfying the :func:`enumerate` protocol over a
        subset or possibly all items contained in the current pith as configured
        by the beartype configuration of the passed violation cause.

        That configuration configures how many items this iterator enumerates
        over. In particular, if that configuration enables:

        * The default :math:`O1` constant-time type-checking strategy (i.e., if
          ``conf.strategy is beartype.BeartypeStrategy.O1``), this iterator
          efficiently enumerates over only a fixed number of (typically only one
          or two) items of this pith.
        * The :math:`On` linear-time type-checking strategy (i.e., if
          ``conf.strategy is beartype.BeartypeStrategy.On``), this iterator
          inefficiently enumerates over *all* items of this pith.

        Parameters
        ----------
        cause: ViolationCause
            Type-checking violation cause finder to be inspected.

        Returns
        -------
        Enumerator
            Iterator yielding zero or more 2-tuples of the standard form
            ``(item_index, item)``, where:

            * ``item_index`` is the 0-based index of the currently enumerated
              item.
            * ``item`` is an arbitrary item of this pith.
        '''

        # Iterator to be returned.
        container_enumerator: Enumerator = None  # type: ignore[assignment]

        # If the only a single item of this container was type-checked by the
        # parent @beartype-generated wrapper function in O(1) time, type-check
        # only the same single item of this container in O(1) time as well.
        if cause.conf.strategy is BeartypeStrategy.O1:
            # 2-tuple of the index and value of an arbitrary item in the same
            # order as the 2-tuples returned by the enumerate() builtin.
            container_enumerator_item = self._get_cause_enumerator_item(cause)

            # Iterator yielding only this 2-tuple.
            container_enumerator = iter((container_enumerator_item,))
        # Else, *ALL* items of this container were type-checked by the parent
        # @beartype-generated wrapper function in O(n) time. In this case,
        # type-check *ALL* items of this container in O(n) time as well.
        else:
            # Iterator yielding all indices and items of this container.
            container_enumerator = enumerate(cause.pith)

        # Return this iterator.
        return container_enumerator

    # ..................{ ABSTRACT                           }..................
    @abstractmethod
    def make_code(
        self, hints_meta: HintsMeta, hint_child_sane: HintSane) -> None:
        '''
        Python expression deeply type-checking the current pith against the
        currently visited container hint described by the passed parameters.

        Parameters
        ----------
        hints_meta : HintsMeta
            Stack of metadata describing all visitable hints currently
            discovered by this breadth-first search (BFS).
        hint_child_sane : HintSane
            **Sanified child hint metadata** (i.e., :data:`.HintSane` object)
            encapsulating the sanification of this child hint to be
            type-checked.
        '''

        pass

# ....................{ SUBCLASSES ~ iterable              }....................
class HintLogicQuasiiterable(HintLogicABC):
    '''
    **Single-argument quasi-iterable hint logic** (i.e., dataclass encapsulating
    all low-level Python code snippets and associated metadata required to
    dynamically generate high-level Python code snippets fully type-checking
    some kind of :pep:`484`- or :pep:`585`-compliant quasi-iterable type hint
    uniquely identified by a common sign, satisfying at least the
    :class:`collections.abc.Iterable` protocol subscripted by exactly one child
    type hint constraining *all* items contained in that container) subclass.

    This logic generates type-checking code for potentially "unsafe" containers
    that are *not* necessarily guaranteed to be safely reiterable. Canonical
    examples would be exhaustible one-time-only containers like generators and
    iterators over "safe" containers that *are* safely reiterable.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self) -> None:
        '''
        Initialize this hint logic.
        '''

        # Initialize our superclass.
        super().__init__(
            get_cause_enumerator_item=_get_cause_enumerator_item_collection,
            # Code snippets dynamically generated by this logic require
            # pseudo-random integers to type-check random sequence items for the
            # proper subset of quasi-iterables that are actually sequences.
            is_var_random_int_needed=True,
        )

    # ..................{ FACTORIES                          }..................
    def make_code(
        self, hints_meta: HintsMeta, hint_child_sane: HintSane) -> None:
        assert isinstance(hints_meta, HintsMeta), (
            f'{repr(hints_meta)} not "HintsMeta" object.')

        # Python expression evaluating to the "collections.abc.Collection" ABC
        # as a hidden parameter passed to the current wrapper function.
        collection_abc_expr = hints_meta.add_func_scope_type_or_types(
            CollectionABC)

        # Python expression evaluating to the "collections.abc.Sequence" ABC as
        # a hidden parameter passed to the current wrapper function.
        sequence_abc_expr = hints_meta.add_func_scope_type_or_types(
            SequenceABC)

        # Increment the integer suffixing the name of a unique local variable
        # storing the value of this child pith *BEFORE* defining this variable.
        hints_meta.pith_curr_var_name_index += 1

        # Name of this local variable.
        pith_child_var_name = PITH_INDEX_TO_VAR_NAME[
            hints_meta.pith_curr_var_name_index]

        # Python expression deeply type-checking this pith against this hint.
        hints_meta.func_curr_code = CODE_PEP484585_QUASIITERABLE_format(
            hint_curr_expr=hints_meta.hint_curr_expr,
            indent_curr=hints_meta.indent_curr,
            pith_curr_assign_expr=hints_meta.pith_curr_assign_expr,
            pith_curr_var_name=hints_meta.pith_curr_var_name,
            pith_child_var_name=pith_child_var_name,
            collection_abc_expr=collection_abc_expr,
            sequence_abc_expr=sequence_abc_expr,
            hint_child_placeholder=hints_meta.enqueue_hint_child_sane(
                hint_sane=hint_child_sane, pith_expr=pith_child_var_name),
        )

# ....................{ SUBCLASSES ~ (reiterable|sequence) }....................
class _HintLogicReiterableOrSequence(HintLogicABC):
    '''
    **Single-argument container hint logic** (i.e., dataclass encapsulating
    all low-level Python code snippets and associated metadata required to
    dynamically generate high-level Python code snippets fully type-checking
    some kind of :pep:`484`- or :pep:`585`-compliant container type hint
    uniquely identified by a common sign, satisfying at least the
    :class:`collections.abc.Container` protocol subscripted by exactly one child
    type hint constraining *all* items contained in that container) subclass.

    Attributes
    ----------
    _pith_child_expr_format : CallableStrFormat
        :meth:`str.format` method bound to a Python expression efficiently
        yielding the value of the next item (which will then be type-checked)
        contained in the **current pith** (which is the parent container
        currently being type-checked). This snippet is expected to contain
        exactly these format variables:

        * ``{pith_curr_var_name}``, expanding to the name of the local variable
          whose value is the current pith.
    '''

    # ..................{ CLASS VARIABLES                    }..................
    # Slot all instance variables defined on this object to minimize the time
    # complexity of both reading and writing variables across frequently called
    # cache dunder methods. Slotting has been shown to reduce read and write
    # costs by approximately ~10%, which is non-trivial.
    __slots__ = (
        '_pith_child_expr_format',
    )

    # Squelch false negatives from mypy. This is absurd. This is mypy. See:
    #     https://github.com/python/mypy/issues/5941
    if TYPE_CHECKING:
        _pith_child_expr_format: CallableStrFormat

    # ..................{ INITIALIZERS                       }..................
    def __init__(
        self,
        pith_child_expr_format: CallableStrFormat,
        **kwargs
    ) -> None:
        '''
        Initialize this hint logic.

        Parameters
        ----------
        See the class docstring for further details. All remaining passed
        keyword parameters are passed as is to the superclass
        :meth:`HintLogicABC.__init__` method.
        '''
        assert callable(pith_child_expr_format), (
            f'{repr(pith_child_expr_format)} uncallable.')

        # Initialize our superclass.
        super().__init__(**kwargs)

        # Classify all passed parameters.
        self._pith_child_expr_format = pith_child_expr_format

    # ..................{ FACTORIES                          }..................
    def make_code(
        self, hints_meta: HintsMeta, hint_child_sane: HintSane) -> None:
        assert isinstance(hints_meta, HintsMeta), (
            f'{repr(hints_meta)} not "HintsMeta" object.')

        # Python expression deeply type-checking this pith against this hint.
        hints_meta.func_curr_code = (
            CODE_PEP484585_REITERABLE_OR_SEQUENCE_format(
                hint_curr_expr=hints_meta.hint_curr_expr,
                indent_curr=hints_meta.indent_curr,
                pith_curr_assign_expr=hints_meta.pith_curr_assign_expr,
                pith_curr_var_name=hints_meta.pith_curr_var_name,
                hint_child_placeholder=hints_meta.enqueue_hint_child_sane(
                    hint_sane=hint_child_sane,
                    # Python expression efficiently yielding some item of this
                    # pith to be deeply type-checked against this child hint.
                    pith_expr=self._pith_child_expr_format(
                        pith_curr_var_name=hints_meta.pith_curr_var_name),
                ),
            )
        )


class HintLogicReiterable(_HintLogicReiterableOrSequence):
    '''
    **Single-argument reiterable hint logic** (i.e., dataclass
    encapsulating all low-level Python code snippets and associated metadata
    required to dynamically generate high-level Python code snippets fully
    type-checking some kind of :pep:`484`- or :pep:`585`-compliant reiterable
    type hint uniquely identified by a common sign, satisfying at least the
    :class:`collections.abc.Collection` protocol subscripted by exactly one
    child type hint constraining *all* items contained in that reiterable)
    subclass.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self) -> None:
        '''
        Initialize this hint logic.
        '''

        # Initialize our superclass.
        super().__init__(
            get_cause_enumerator_item=_get_cause_enumerator_item_reiterable,
            pith_child_expr_format=(
                CODE_PEP484585_REITERABLE_PITH_CHILD_EXPR_format),
        )


class HintLogicSequence(_HintLogicReiterableOrSequence):
    '''
    **Single-argument sequence hint logic** (i.e., dataclass encapsulating
    all low-level Python code snippets and associated metadata required to
    dynamically generate high-level Python code snippets fully type-checking
    some kind of :pep:`484`- or :pep:`585`-compliant sequence type hint
    uniquely identified by a common sign, satisfying at least the
    :class:`collections.abc.Sequence` protocol subscripted by exactly one child
    type hint constraining *all* items contained in that sequence) subclass.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self) -> None:
        '''
        Initialize this hint logic.
        '''

        # Initialize our superclass.
        super().__init__(
            get_cause_enumerator_item=_get_cause_enumerator_item_sequence,
            # Code snippets dynamically generated by this logic require
            # pseudo-random integers to type-check random sequence items.
            is_var_random_int_needed=True,
            pith_child_expr_format=(
                CODE_PEP484585_SEQUENCE_PITH_CHILD_EXPR_format),
        )

# ..................{ PRIVATE ~ getters                      }..................
#FIXME: Shift these into a new private utility class. *shrug*

def _get_cause_enumerator_item_collection(
    cause: ViolationCause) -> EnumeratorItem:
    '''
    2-tuple ``(item_index, item)`` describing the first item of the passed
    collection satisfying the format of the :func:`enumerate` iterator.

    Parameters
    ----------
    cause: ViolationCause
        Type-checking violation cause finder to be inspected.

    Returns
    -------
    EnumeratorItem
        2-tuple of the standard form ``(item_index, item)`` returned by the
        :func:`enumerate` builtin, where:

        * ``item_index`` is the 0-based index of the first item of this
          collection.
        * ``item`` is that item.
    '''
    assert isinstance(cause.pith, CollectionABC), (
        f'Violation cause {repr(cause)} pith not collection.')

    # Return either...
    return (
        # If this cause describes a sequence, a pseudo-random item of this
        # sequence;
        _get_cause_enumerator_item_sequence(cause)
        if isinstance(cause.pith, SequenceABC) else
        # Else, this cause does *NOT* describe a sequence. Since this cause
        # describes a collection, this cause *MUST* necessarily describe a
        # reiterable by elimination. In this case, the first item of this
        # reiterable.
        _get_cause_enumerator_item_reiterable(cause)
    )


def _get_cause_enumerator_item_reiterable(
    cause: ViolationCause) -> EnumeratorItem:
    '''
    2-tuple ``(item_index, item)`` describing the first item of the passed
    reiterable satisfying the format of the :func:`enumerate` iterator.

    Parameters
    ----------
    cause: ViolationCause
        Type-checking violation cause finder to be inspected.

    Returns
    -------
    EnumeratorItem
        2-tuple of the standard form ``(item_index, item)`` returned by the
        :func:`enumerate` builtin, where:

        * ``item_index`` is the 0-based index of the first item of this
          reiterable.
        * ``item`` is that item.
    '''

    # First item of this container.
    item = next(iter(cause.pith))

    # 0-based index of this item for readability purposes.
    item_index = 0

    # Return a 2-tuple "(item_index, item)" describing this item.
    return (item_index, item)


def _get_cause_enumerator_item_sequence(
    cause: ViolationCause) -> EnumeratorItem:
    '''
    2-tuple ``(item_index, item)`` describing a pseudo-random item of the passed
    sequence satisfying the format of the :func:`enumerate` iterator.

    Parameters
    ----------
    cause: ViolationCause
        Type-checking violation cause finder to be inspected.

    Returns
    -------
    EnumeratorItem
        2-tuple of the standard form ``(item_index, item)`` returned by the
        :func:`enumerate` builtin, where:

        * ``item_index`` is the 0-based index of a pseudo-random item of this
          sequence.
        * ``item`` is that item.
    '''

    assert cause.random_int is not None, (
        f'Violation cause {repr(cause)} pseudo-random integer is "None".')

    # 0-based index of this item calculated from this random integer in the
    # *SAME EXACT WAY* as in the parent @beartype-generated wrapper.
    item_index = cause.random_int % len(cause.pith)

    # Pseudo-random item with this index in this sequence.
    item = cause.pith[item_index]

    # Return a 2-tuple "(item_index, item)" describing this item.
    return (item_index, item)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype decorator exception unit tests.**

This submodule unit tests the :func:`beartype.beartype` decorator with respect
to exceptions raised by the :mod:`beartype._check.error.errmain` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                              }....................
def test_decor_violation_culprits() -> None:
    '''
    Test the :func:`beartype.beartype` decorator with respect to **type-checking
    violations** (i.e., :exc:`beartype.roar.BeartypeCallHintViolation`
    exceptions raised by callables decorated by that decorator).
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import BeartypeCallHintViolation
    from beartype.typing import List
    from pytest import raises

    # ....................{ ASSERTS ~ strong ref           }....................
    class SpiritBearIGiveYouSalmonToGoAway(object):
        '''
        Arbitrary user-defined class.
        '''

        pass

    class SpiritBearIGiftYouHoneyNotToStay(object):
        '''
        Arbitrary user-defined class.
        '''

        pass

    # Arbitrary object guaranteed to *NOT* be an instance of this class.
    #
    # Since CPython permits weak references to almost all instances of
    # user-defined classes (including both the above class and the "object"
    # superclass), this object exhibits the expected behaviour from the
    # "culprits" API exercised below that depends upon weak references.
    SPIRIT_BEAR_REFUSE_TO_GO_AWAY = SpiritBearIGiftYouHoneyNotToStay()

    @beartype
    def when_spirit_bear_hibernates_in_your_bed(
        best_bear_den: SpiritBearIGiveYouSalmonToGoAway) -> None:
        '''
        Arbitrary function decorated by :func:`beartype.beartype` annotated by a
        type hint inducing edge-case behaviour in type-checking violations
        raised by this callable.
        '''

        pass

    # Assert that calling this function with a parameter violating its type hint
    # raises the expected exception.
    with raises(BeartypeCallHintViolation) as exception_info:
        when_spirit_bear_hibernates_in_your_bed(SPIRIT_BEAR_REFUSE_TO_GO_AWAY)

    # Exception captured by the prior call to this wrapper function.
    exception = exception_info.value

    # Assert that exactly one culprit was responsible for this exception.
    assert len(exception.culprits) == 1

    # Culprit of this exception.
    culprit = exception.culprits[0]

    # Assert that this culprit is the same object passed to this function.
    assert culprit is SPIRIT_BEAR_REFUSE_TO_GO_AWAY

    # ....................{ ASSERTS ~ weak ref             }....................
    # List of lists of byte strings violating type hints annotating the
    # we_are_all_spirit_bear() function defined below.
    #
    # Since CPython prohibits weak references to both lists *AND* byte strings,
    # this data structure exhibits known flaws in the "culprits" API exercised
    # below that depends upon weak references.
    YOU_WENT_THERE_SPIRIT_BEAR = (
        b'Why do you sleep in my pinball room, Spirit Bear?')
    DO_NOT_GO_THERE_SPIRIT_BEAR = [[YOU_WENT_THERE_SPIRIT_BEAR]]

    @beartype
    def we_are_all_spirit_bear(
        best_bear_dens: List[List[str]]) -> None:
        '''
        Arbitrary function decorated by :func:`beartype.beartype` annotated by a
        type hint inducing edge-case behaviour in type-checking violations
        raised by this callable.
        '''

        pass

    # Assert that calling this function with a parameter violating its type hint
    # raises the expected exception.
    with raises(BeartypeCallHintViolation) as exception_info:
        we_are_all_spirit_bear(DO_NOT_GO_THERE_SPIRIT_BEAR)

    # Exception captured by the prior call to this wrapper function.
    exception = exception_info.value

    # Assert that exactly two culprits were responsible for this exception.
    assert len(exception.culprits) == 2

    # Root and leaf culprits of this exception.
    root_culprit = exception.culprits[0]
    leaf_culprit = exception.culprits[1]

    # Assert that both of these culprits are the machine-readable
    # representations of their underlying objects rather than those objects.
    # Why? Because CPython prohibits weak references to those objects. *sigh*
    assert isinstance(root_culprit, str)
    assert isinstance(leaf_culprit, str)
    assert root_culprit == repr(DO_NOT_GO_THERE_SPIRIT_BEAR)
    assert leaf_culprit == repr(YOU_WENT_THERE_SPIRIT_BEAR)


def test_decor_violation_pickle() -> None:
    '''
    Test the :func:`beartype.beartype` decorator with respect to **type-checking
    violation pickling** (i.e., (de)serializing
    :exc:`beartype.roar.BeartypeCallHintViolation` exceptions raised by
    callables decorated by that decorator via the :mod:`pickle` module).
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from pickle import dumps, loads
    from beartype.door import die_if_unbearable
    from beartype.roar import BeartypeDoorHintViolation

    # ....................{ CLASSES                        }....................
    class TheHorridTruth(object):
        '''
        Arbitrary user-defined type.

        Note that this unit test is deterministic *only* when the
        :func:`.die_if_unbearable` function is passed an instance of a
        user-defined types. When that function is instead passed a builtin
        scalar (e.g., string, integer), this unit test becomes
        non-deterministic. Why? Object interning, probably. For efficiency,
        Python **interns** (i.e., internalizes as globally cached objects)
        simple primitives like builtin scalars. How and when Python does so
        probably varies depending on contextual runtime features like the
        current platform, CPython compiler options and *whatevahs*.
        '''

        pass

    # Arbitrary instance of this type.
    the_awful_facts = TheHorridTruth()

    # ....................{ ASSERTS                        }....................
    # Trivially raise a type-checking violation.
    try:
        die_if_unbearable(the_awful_facts, int)
    except BeartypeDoorHintViolation as exception:
        # Pickled string embodying this exception's state.
        exception_dumped = dumps(exception)

        # Unpickled exception resembling the original exception.
        exception_loaded = loads(exception_dumped)

        # Assert that this unpickled exception is of the same type.
        assert isinstance(exception_loaded, BeartypeDoorHintViolation)

        # Assert that both the original *AND* unpickled exception provide the
        # expected number of culprits.
        assert len(exception.culprits) == 1
        assert len(exception_loaded.culprits) == 1

        # Assert that the unpickled exception stores *ONLY* the machine-readable 
        # representation of this culprit. Why? Because the weak reference to the
        # culprit stored by the original exception is *NOT* pickleable.
        assert repr(exception.culprits[0]) == exception_loaded.culprits[0]


def test_decor_violation_types() -> None:
    '''
    Test the
    :func:`beartype.beartype` decorator with respect to the
    :attr:`beartype.BeartypeConf.violation_param_type` and
    :attr:`beartype.BeartypeConf.violation_return_type` configuration
    parameters.
    '''

    # ..................{ IMPORTS                            }..................
    # Defer test-specific imports.
    from beartype import (
        BeartypeConf,
        beartype,
    )
    from beartype.typing import (
        List,
        Tuple,
        Union,
    )
    from pytest import (
        raises,
        warns,
    )

    # ..................{ CLASSES                            }..................
    class TheVacantWoods(Exception):
        '''
        Arbitrary exception subclass.
        '''

        pass


    class SpreadRoundHim(Exception):
        '''
        Arbitrary exception subclass.
        '''

        pass


    class LowInTheWest(UserWarning):
        '''
        Arbitrary warning subclass.
        '''

        pass


    class WhereHeStood(UserWarning):
        '''
        Arbitrary warning subclass.
        '''

        pass

    # ..................{ CALLABLES                          }..................
    def the_clear_and(
        garish_hills: List[str], the_distinct_valley) -> (
        Union[int, Tuple[str, ...]]):
        '''
        Arbitrary callable to be decorated by the :func:`beartype.beartype`
        decorator, configured in various ways below.
        '''

        return the_distinct_valley

    # Beartype decorator configured to raise different types of user-defined
    # exceptions on invalid parameters and returns.
    beartype_raise = beartype(conf=BeartypeConf(
        violation_param_type=TheVacantWoods,
        violation_return_type=SpreadRoundHim,
    ))

    # Beartype decorator configured to emit different types of user-defined
    # warnings on invalid parameters and returns.
    beartype_warn = beartype(conf=BeartypeConf(
        violation_param_type=LowInTheWest,
        violation_return_type=WhereHeStood,
        # is_debug=True,
    ))

    # Above callable decorated with type-checking raising custom exceptions.
    the_clear_and_raise = beartype_raise(the_clear_and)

    # Above callable decorated with type-checking emitting custom warnings.
    the_clear_and_warn = beartype_warn(the_clear_and)

    # ..................{ CONSTANTS                          }..................
    # Arbitrary list of strings to be passed to these callables.
    WHITHER_HAVE_FLED = [
        'Low in the west, the clear and garish hills,',
        'The distinct valley and the vacant woods,',
    ]

    # Arbitrary tuple of strings to be passed to these callables.
    THE_HUES_OF_HEAVEN = (
        'Spread round him where he stood. Whither have fled',
        'The hues of heaven that canopied his bower',
    )

    # ..................{ PASS                               }..................
    # Assert that these decorated callables neither raise exceptions nor emit
    # warnings when passed valid parameters.
    assert the_clear_and_raise(WHITHER_HAVE_FLED, THE_HUES_OF_HEAVEN) is (
        THE_HUES_OF_HEAVEN)
    assert the_clear_and_warn(WHITHER_HAVE_FLED, THE_HUES_OF_HEAVEN) is (
        THE_HUES_OF_HEAVEN)

    # Assert that this decorated callable raises the expected exception that
    # this callable was configured to raise on receiving an invalid parameter.
    with raises(TheVacantWoods):
        the_clear_and_raise(THE_HUES_OF_HEAVEN, WHITHER_HAVE_FLED)

    # Assert that this decorated callable raises the expected exception that
    # this callable was configured to raise on returning an invalid return.
    with raises(SpreadRoundHim):
        the_clear_and_raise(WHITHER_HAVE_FLED, WHITHER_HAVE_FLED)

    # Assert that this decorated callable emits the expected warning that
    # this callable was configured to emit on receiving an invalid parameter.
    with warns(LowInTheWest):
        the_clear_and_warn(THE_HUES_OF_HEAVEN, THE_HUES_OF_HEAVEN)

    # Assert that this decorated callable emits the expected warning that
    # this callable was configured to emit on returning an invalid return.
    with warns(WhereHeStood):
        the_clear_and_warn(WHITHER_HAVE_FLED, WHITHER_HAVE_FLED)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **sequence factory** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.kind.sequence.utilseqmake` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ updaters                   }....................
def test_make_stack() -> None:
    '''
    Test the :func:`beartype._util.kind.sequence.utilseqmake.make_stack`
    function.
    '''

    # Defer test-specific imports.
    from beartype._util.kind.sequence.utilseqmake import make_stack

    # Assert that this factory creates the expected stack from empty iterables,
    # including both empty tuples and lists.
    assert make_stack(()) == make_stack([]) == []

    # Assert that this factory creates the expected stack from non-empty
    # iterables, including both non-empty tuples and lists.
    assert (
        make_stack(('Wan', 'moonlight', 'even', 'to', 'fulness:', 'not', 'a', 'star'))) == (
        make_stack(['Wan', 'moonlight', 'even', 'to', 'fulness:', 'not', 'a', 'star'])) == (
        ['star', 'a', 'not', 'fulness:', 'to', 'even', 'moonlight', 'Wan'])



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`695`-compliant **type alias** (i.e., objects created via the
``type`` statement under Python >= 3.12) utilities.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ TODO                               }....................
#FIXME: Consider generalizing @beartype's PEP 695 implementation to additionally
#support local type aliases (i.e., defined in the local scope of a callable
#rather than at global scope) containing one or more unquoted forward
#references. Currently, @beartype intentionally fails to support this: e.g.,
#    type global_alias = ...
#    die_if_unbearable('lolwut', global_alias)  # <-- this is fine
#
#    def muh_func(...) -> ...:
#        type local_alias = ...
#        die_if_unbearable('lolwut', local_alias)  # <-- raises an exception
#
#The reasons are obscure. Very well. CPython's current implementation of local
#type aliases is probably very buggy. An upstream issue describing this
#bugginess should be submitted. When doing so, please publicly declare that PEP
#695 appears to have been poorly tested. As evidence, note that PEP 695 itself
#advises use of the following idiom:
#    # A type alias that includes a forward reference
#    type AnimalOrVegetable = Animal | "Vegetable"
#
#*THAT DOES NOT ACTUALLY WORK AT RUNTIME.* Nobody tested that. This is why I
#facepalm. Notably, PEP 604-compliant new-style unions prohibit strings. They
#probably shouldn't, but they've *ALWAYS* behaved that way, and nobody's updated
#them to behave more intelligently -- probably because doing so would require
#updating the isinstance() builtin (which also accepts PEP 604-compliant
#new-style unions) to behave more intelligiently and ain't nobody goin' there:
#e.g.,
#
#    $ python3.12
#    >>> type AnimalOrVegetable = "Animal" | "Vegetable"
#    >>> AnimalOrVegetable.__value__
#    Traceback (most recent call last):
#      Cell In[3], line 1
#        AnimalOrVegetable.__value__
#      Cell In[2], line 1 in AnimalOrVegetable
#        type AnimalOrVegetable = "Animal" | "Vegetable"
#    TypeError: unsupported operand type(s) for |: 'str' and 'str'
#
#For further details, see the comment below prefixed by:
#           # If that module fails to define this alias as a global variable,
#
#Since CPython is unlikely to resolve its bugginess anytime soon, it inevitably
#falls to @beartype to resolve this. Thankfully, @beartype *CAN* resolve this.
#Unthankfully, doing so will require @beartype to implement a new PEP
#695-specific AST transform from the "beartype.claw" subpackage augmenting *ALL*
#PEP 695-compliant local type aliases (so, probably *ALL* type aliases
#regardless of scope for simplicity) as follows:
#    # "beartype.claw" should transform this...
#    type {alias_name} = {alias_value}
#
#    # ...into this.
#    from beartype._util.hint.pep.proposal.pep695 import (
#        iter_hint_pep695_unsubbed_forwardrefs as
#        __iter_hint_pep695_forwardref_beartype__
#    )
#    type {alias_name} = {alias_value}
#    for __hint_pep695_forwardref_beartype__ in (
#        __iter_hint_pep695_forwardref_beartype__({alias_name})):
#        # If the current scope is module scope, prefer an efficient
#        # non-exec()-based solution. Note that this optimization does *NOT*
#        # generalize to other scopes, for obscure reasons delineated here:
#        #     https://stackoverflow.com/a/8028772/2809027
#        if globals() is locals():
#            globals()[__hint_pep695_forwardref_beartype__.__name_beartype__] =
#                __hint_pep695_forwardref_beartype__)
#        # Else, the current scope is *NOT* module scope. In this case,
#        # fallback to an inefficient exec()-based solution.
#        else:
#            exec(f'{__hint_pep695_forwardref_beartype__.__name_beartype__} = __hint_pep695_forwardref_beartype__')
#
#    #FIXME: Technically, this *ONLY* needs to be done if the
#    #iter_hint_pep695_unsubbed_forwardrefs() iterator returned something. *shrug*
#    # Intentionally redefine this alias. Although this appears to be an
#    # inefficient noop, this is in fact an essential operation. Why?
#    # Because the prior successful access of the "__value__" dunder
#    # variable silently cached and thus froze the value of this alias.
#    # However, alias values are *NOT* necessarily safely freezable at
#    # alias definition time. The canonical example of alias values that
#    # are *NOT* safely freezable at alias definition time are mutually
#    # recursive aliases (i.e., aliases whose values circularly refer to
#    # one another): e.g.,
#    #     type a = b
#    #     type b = a
#    #
#    # PEP 695 provides no explicit means of uncaching alias values.
#    # Our only recourse is to repetitiously redefine this alias.
#    type {alias_name} = {alias_value}
#
#We're currently unclear whether anyone actually cares about this. Ergo, we
#adopted the quick-and-dirty approach of raising exceptions instead. Yikes!

# ....................{ IMPORTS                            }....................
from beartype.meta import URL_ISSUES
from beartype.roar import BeartypeDecorHintPep695Exception
from beartype.typing import (
    Iterable,
    Optional,
)
from beartype._cave._cavefast import (
    # HintGenericSubscriptedType,
    HintPep695TypeAlias,
    Pep695ParameterizableTypes,
)
from beartype._check.forward.reference.fwdrefmake import (
    make_forwardref_indexable_subtype)
from beartype._check.forward.reference.fwdrefmeta import BeartypeForwardRefMeta
from beartype._data.typing.datatypingport import Hint
from beartype._data.typing.datatyping import (
    LexicalScope,
    Pep695Parameterizable,
    TuplePep484612646TypeArgsPacked,
)
from beartype._util.error.utilerrget import get_name_error_attr_name
from beartype._util.module.utilmodget import get_module_imported_or_none
from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_12
from beartype._data.kind.datakindiota import SENTINEL

# ....................{ TESTERS                            }....................
def is_hint_pep695_subbed(hint: Hint) -> bool:
    '''
    :data:`True` only if the passed type hint is a :pep:`695`-compliant
    **subscripted type alias** (i.e., object created by subscripting an object
    created by a statement of the form ``type {alias_name}[{type_var}] =
    {alias_value}`` by one or more child type hints) is ignorable.

    This tester disambiguates between **unrecognized subscripted builtin type
    hints** (i.e., C-based type hints instantiated by subscripting pure-Python
    origin types unrecognized by :mod:`beartype` and thus PEP-noncompliant)
    from subscripted type aliases (which are clearly :pep:`695`-compliant).
    Superficially, subscripted type aliases resemble unrecognized subscripted
    builtin type hints due to internally reusing the same low-level
    :pep:`585`-compliant :class:`types.GenericAlias` architecture. We sigh.

    This tester is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as the implementation efficiently reduces to
    a one-liner.

    Caveats
    -------
    Note that, although subscriptable type aliases superficially appear to be
    "pre-subscripted" by :pep:`484`-compliant type variables, this
    "pre-subscription" is simply syntactic sugar; subscriptable type aliases
    remain unsubscripted until explicitly subscripted by concrete types: e.g.,

    .. code-block:: pycon

       >>> from beartype._util.hint.pep.proposal.pep695 import (
       ...     is_hint_pep695_subbed)
       >>> type subscriptable_alias[T] = int | T
       >>> is_hint_pep695_subbed(subscriptable_alias)
       False
       >>> is_hint_pep695_subbed(subscriptable_alias[float])
       True

    Parameters
    ----------
    hint : Hint
        Type hint to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this hint is a subscripted type alias.
    '''

    # Avoid circular import dependencies.
    from beartype._util.hint.pep.utilpepget import get_hint_pep_origin_or_none
    # from beartype._util.hint.pep.proposal.pep585 import (
    #     is_hint_pep585_builtin_subbed)

    #FIXME: Actually, this test is surprisingly computationally expensive.
    #Ignore for now as the test below currently suffices. *shrug*
    # # If this hint is a PEP 585-compliant subscripted builtin hint, immediately
    # # return false. All PEP 695-compliant subscripted type aliases are
    # # implemented as PEP 585-compliant subscripted builtin hints, interestingly.
    # if not is_hint_pep585_builtin_subbed(hint):
    #     return False
    # # Else, this hint is a PEP 585-compliant subscripted builtin hint and thus
    # # *COULD* be a PEP 695-compliant subscripted type alias. Further detection
    # # is warranted.

    # Origin (i.e., value of the "__origin__" dunder attribute) originating this
    # hint if this hint originates from such a type *OR* "None" otherwise (i.e.,
    # if this hint originates from such a type).
    hint_origin = get_hint_pep_origin_or_none(hint)

    # Return true only if this origin is a PEP 695-compliant unsubscripted type
    # alias. Yes. It really is this non-trivial, folks. *sigh*
    return isinstance(hint_origin, HintPep695TypeAlias)

# ....................{ GETTERS                            }....................
#FIXME: Unit test us up, please.
def get_hint_pep695_unsubbed_alias(
    # Mandatory parameters.
    hint: HintPep695TypeAlias,

    # Optional parameters.
    exception_prefix: str = '',
) -> Hint:
    '''
    **Non-alias type hint** (i.e., type hint that is *not* a
    :pep:`695`-compliant type alias) encapsulated by the passed
    :pep:`695`-compliant **unsubscripted type alias** (i.e., object created by a
    statement of the form ``type {alias_name} = {alias_value}``).

    This getter is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), for subtle reasons pertaining to unquoted
    forward references. Notably, memoizing this getter would prevent the
    external caller of the higher-level
    :func:`.iter_hint_pep695_unsubbed_forwardrefs` iterator calling this
    lower-level getter from externally modifying this type alias by forcefully
    injecting forward reference proxies into this alias.

    Parameters
    ----------
    hint : HintPep695TypeAlias
        Unsubscripted type alias to be inspected.
    exception_prefix : str, default: ''
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Returns
    -------
    Hint
        Unaliased type hint encapsulated by this type alias.

    Raises
    ------
    BeartypeDecorHintPep695Exception
        If this type hint is *not* an unsubscripted type alias.
    NameError
        If this unsubscripted type alias contains one or more **unquoted forward
        references** to undefined types.
    '''

    # If this hint is *NOT* a PEP 695-compliant unsubscripted type alias, raise
    # an exception.
    if not isinstance(hint, HintPep695TypeAlias):
        raise BeartypeDecorHintPep695Exception(
            f'{exception_prefix}type hint {repr(hint)} '
            f'not PEP 695 unsubscripted type alias.'
        )
    # Else, this hint is a PEP 695-compliant unsubscripted type alias.

    # While the Universe continues infinitely expanding...
    while True:
        # Reduce this type alias to the type hint aliased by this alias, which
        # itself is possibly a nested type alias. Oh, it happens.
        #
        # Note that doing so implicitly raises a "NameError" if this alias
        # contains one or more unquoted forward references to undefined types.
        hint = hint.__value__  # type: ignore[attr-defined]

        # If this type hint is *NOT* a nested type alias, break this iteration.
        if not isinstance(hint, HintPep695TypeAlias):
            break
        # Else, this type hint is a nested type alias. In this case, continue
        # iteratively unwrapping this nested type alias.

    # Return this unaliased type alias.
    return hint

# ....................{ ADDERS                             }....................
#FIXME: Unit test us up, please.
def add_func_scope_hint_pep695_parameterizable_typeparams(
    # Mandatory parameters.
    func_scope: LexicalScope,
    parameterizable: Pep695Parameterizable,

    # Optional parameters.
    exception_prefix: str = '',
) -> None:
    '''
    Add one key-value pair mapping the name to value of each **type parameter**
    (i.e., :pep:`484`-compliant type variable, pep:`612`-compliant parameter
    specification, or :pep:`646`-compliant type variable tuple) implicitly
    instantiated with :pep:`695`-compliant type parameter syntax parametrizing
    the **parameterizable** (i.e., pure-Python class, pure-Python function, or
    :pep:`695`-compliant type alias) described by the passed lexical scope.

    Caveats
    -------
    This function silently overrides existing key-value pairs sharing the same
    names as type parameters parametrizing this parameterizable. Why? Because
    doing so mimics lexical scoping implemented by the CPython parser itself,
    which this function effectively masquerades as.

    Parameters
    ----------
    func_scope : LexicalScope
        Local or global scope to map these type parameters to.
    parameterizable : Pep695Parameterizable
        :pep:`695`-compliant parameterizable to be inspected.
    exception_prefix : str, optional
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Raises
    ------
    BeartypeDecorHintPep695Exception
        If either:

        * This object is *not* a :pep:`695`-compliant parameterizable.
        * This object is a :pep:`695`-compliant parameterizable parametrized by
          a type parameter already parametrizing a **parent lexical scope**
          (e.g., class) of this parameterizable. By :pep:`695`, type parameters
          are *not* safely reusable across nested lexical scopes:

              Consistent with the scoping rules defined in PEP 484, type
              checkers should generate an error if inner-scoped generic classes,
              functions, or type aliases reuse the same type parameter name as
              an outer scope.
    '''
    assert isinstance(func_scope, dict), f'{repr(func_scope)} not dictionary.'
    # print(f'Updating PEP 695 parameterizable {repr(parameterizable)} scope {repr(func_scope)}...')

    # ....................{ IMPORTS                        }....................
    # Avoid circular import dependencies.
    from beartype._util.hint.pep.proposal.pep484612646 import (
        get_hint_pep484612646_typearg_packed_name,
        is_hint_pep484612646_typearg_packed,
    )

    # ....................{ LOCALS                         }....................
    # Tuple of all type parameters parametrizing this parameterizable.
    typeparams = _get_hint_pep695_parameterizable_typeparams(
        parameterizable=parameterizable, exception_prefix=exception_prefix)
    # print(f'Adding PEP 695 parameterizable {repr(parameterizable)} type parameters {repr(typeparams)}...')

    # ....................{ LOOP                           }....................
    # For each type parameter parametrizing this parameterizable...
    for typeparam in typeparams:
        # Unqualified basename of this type parameter.
        typeparam_name = get_hint_pep484612646_typearg_packed_name(
            hint=typeparam, exception_prefix=exception_prefix)

        # Existing attribute sharing the same name already added to this forward
        # scope by the caller if any *OR* the placeholder sentinel otherwise
        # (i.e., if this scope contains no attribute with this name).
        func_scope_attr = func_scope.get(typeparam_name, SENTINEL)

        # If...
        if (
            # This scope already contains an attribute with this name *AND*...
            func_scope_attr is not SENTINEL and
            # This attribute is also a type parameter...
            is_hint_pep484612646_typearg_packed(func_scope_attr)
        ):
            assert isinstance(exception_prefix, str), (
                f'{repr(exception_prefix)} not string.')

            # Then this previously added type parameter conflicts with this
            # currently iterated type parameter. This common case arises when
            # callers erroneously reuse the name of a type parameter across
            # nested lexical scopes, as explicitly prohibited by PEP 695: e.g.,
            #     class MuhGeneric[T]:       # <-- good
            #         def muh_func[T](): ... # <-- *BAD* according to PEP 695
            raise BeartypeDecorHintPep695Exception(
                f'{exception_prefix}'
                f'nested {repr(parameterizable)} '
                f'PEP 695 type parameter {repr(typeparam)} shadows parent '
                f'PEP 695 type parameter of same name; '
                f'however, PEP 695 prohibits type parameter reuse across '
                f'nested lexical scopes.'
            )
        # Else, either this scope does not already contain an attribute with
        # this name *OR* this scope does but this attribute is not also a type
        # parameter. In either case, this type parameter silently shadows (i.e.,
        # overrides) this attribute. Although attribute shadowing could be
        # regarded as a problematic design pattern indicative of a genuine
        # issue, both CPython and PEP 695 explicitly allow, enable, and arguably
        # encourage attribute shadowing where the attribute being shadowed is
        # itself *NOT* also a type parameter. It's not for @beartype to police
        # valid use cases. That's what static type-checkers do. So, we don't.

        # Map this basename to this type parameter in this lexical scope,
        # silently shadowing any existing attribute of the same name (if any).
        func_scope[typeparam_name] = typeparam


def _get_hint_pep695_parameterizable_typeparams(
    # Mandatory parameters.
    parameterizable: Pep695Parameterizable,

    # Optional parameters.
    exception_prefix: str = '',
) -> TuplePep484612646TypeArgsPacked:
    '''
    Tuple of the zero or more **type parameters** (i.e., :pep:`484`-compliant
    type variables, pep:`612`-compliant parameter specifications, and
    :pep:`646`-compliant type variable tuples) implicitly instantiated with
    :pep:`695`-compliant type parameter syntax parametrizing the passed
    **parameterizable** (i.e., pure-Python class, pure-Python function, or
    :pep:`695`-compliant type alias).

    Parameters
    ----------
    parameterizable : Pep695Parameterizable
        :pep:`695`-compliant parameterizable to be inspected.
    exception_prefix : str, optional
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Returns
    -------
    TuplePep484612646TypeArgsPacked
        Tuple of all type parameters parametrizing this parameterizable.

    Raises
    ------
    BeartypeDecorHintPep695Exception
        If this object is *not* a :pep:`695`-compliant parameterizable.
    '''

    # If this object is *NOT* parameterizable under PEP 695, raise an exception.
    if not isinstance(parameterizable, Pep695ParameterizableTypes):
        raise BeartypeDecorHintPep695Exception(
            f'{exception_prefix}'
            f'{repr(parameterizable)} not PEP 695-parameterizable '
            f'(i.e., neither pure-Python type, pure-Python function, nor '
            f'PEP 695-compliant type alias).'
        )
    # Else, this object is parameterizable under PEP 695.

    # Return either...
    return (
        # If the active Python interpreter targets Python >= 3.12 and thus
        # supports PEP 695, the PEP 695-compliant tuple of all type parameters
        # parametrizing this object.
        parameterizable.__type_params__  # type: ignore[return-value,union-attr]
        if IS_PYTHON_AT_LEAST_3_12 else
        # Else, the active Python interpreter targets Python <= 3.11 and thus
        # fails to support PEP 695. In this case, the empty tuple.
        ()
    )

# ....................{ ITERATORS                          }....................
def iter_hint_pep695_unsubbed_forwardrefs(
    # Mandatory parameters.
    hint: HintPep695TypeAlias,

    # Optional parameters.
    exception_prefix: str = '',
) -> Iterable[BeartypeForwardRefMeta]:
    '''
    Iteratively create and yield one **forward reference proxy** (i.e.,
    :class:`beartype._check.forward.reference.fwdrefabc.BeartypeForwardRefABC`
    subclass) for each unquoted relative forward reference in the passed
    :pep:`695`-compliant **unsubscripted type alias** (i.e., object created by a
    statement of the form ``type {alias_name} = {alias_value}``) to the
    underlying type hint lazily referred to by this type alias.

    This iterator is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as this iterator is intended to be
    called only once per unsubscripted type alias in userspace code dynamically
    instrumented by beartype import hook abstract syntax tree (AST) node
    transformers. Since those transformers are expected to replace *all*
    unquoted relative forward references in this unsubscripted type alias with
    corresponding forward reference proxies, calling this iterator again on any
    unsubscripted type alias instrumented in this way *should* silently reduce
    to a noop. "Should" is doing a lot of heavy lifting here.

    Parameters
    ----------
    hint : HintPep695TypeAlias
        Unsubscripted type alias to be iterated over.
    exception_prefix : str, optional
        Human-readable substring prefixing exception messages raised by this
        reducer. Defaults to the empty string.

    Yields
    ------
    BeartypeForwardRefMeta
        Forward reference proxy encapsulating the next unquoted relative forward
        reference in this unsubscripted type alias.

    Raises
    ------
    BeartypeDecorHintPep695Exception
        If this type hint is *not* an unsubscripted type alias.
    '''

    # Unqualified basename of the previous undeclared attribute in this alias.
    hint_ref_name_prev: Optional[str] = None

    # While this alias still contains one or more forward references to
    # attributes *NOT* defined by the module declaring this alias...
    while True:
        # Attempt to...
        try:
            # print(f'type {hint.__name__} = {hint.__value__}')

            # Reduce this alias to the type hint it lazily refers to. If this
            # alias contains *NO* forward references to undeclared attributes,
            # this reduction *SHOULD* succeed. Let's pretend we mean that.
            #
            # Note that _get_hint_pep695_unsubbed_alias() is memoized and
            # thus intentionally called with positional arguments.
            get_hint_pep695_unsubbed_alias(hint, exception_prefix)

            # This reduction raised *NO* exception and thus succeeded. In this
            # case, immediately halt iteration.
            break
        # If doing so raises a builtin "NameError" exception, this alias
        # contains one or more forward references to undeclared attributes. In
        # this case...
        except NameError as exception:
            # Unqualified basename of this alias (i.e., name of the global or
            # local variable assigned to by the left-hand side of this alias).
            hint_name = repr(hint)

            # Fully-qualified name of the external third-party module defining
            # this alias.
            hint_module_name = hint.__module__
            # print(f'hint_module_name: {hint_module_name}')

            # If this alias defines *NO* module name, raise an exception.
            #
            # Note that this should *NEVER* happen. Nonetheless, static
            # type-checkers like mypy insist this can happen. It almost
            # certainly can't. Nonetheless, let's dot our i's and cross our t's.
            if not hint_module_name:
                raise BeartypeDecorHintPep695Exception(
                    f'{exception_prefix}PEP 695 type alias "{hint_name}" '
                    f'module undefined (i.e., "__module__" attribute '
                    f'either "None" or the empty string).'
                ) from exception
            # Else, this alias defines a module name.

            # That module as its previously imported object.
            hint_module = get_module_imported_or_none(hint_module_name)

            # Unqualified basename of the next remaining undeclared attribute
            # contained in this alias relative to that module.
            hint_ref_name = get_name_error_attr_name(exception)
            # print(f'hint: {hint}; hint_ref_name: {hint_ref_name}')

            # If this attribute is the same as that of the prior iteration of
            # this "while" loop, then that iteration *MUST* have failed to
            # define this attribute as a global variable of that module. In this
            # case, raise an exception.
            #
            # Note that this should *NEVER* happen. Of course, this frequently
            # happens. Specifically, this happens whenever the caller defines a
            # callable defining type alias as a local variable containing one or
            # more unquoted relative forward reference to user-defined classes
            # that have yet to be defined. Why? Because CPython's low-level
            # C-based implementation of PEP 695-compliant type aliases currently
            # fails to properly resolve unquoted relative forward references
            # defined in a local rather than global scope: e.g.,
            #    >>> def foo():
            #    ...     type bar = wut
            #    ...     globals()['wut'] = str
            #    ...     print(bar.__value__)
            #    ...     class wut(object): pass  # <-- causes madness; WTF!?!?
            #    >>> foo()
            #    NameError: cannot access free variable 'wut' where it is not
            #    associated with a value in enclosing scope
            #
            # Why does this matter? Because the abstract syntax tree (AST)
            # transformation implemented by "beartype.claw" import hooks
            # dynamically declares the objects that these forward references
            # refer. Due to deficiencies [read: bugs] in CPython's type alias
            # implementation, local type aliases remain unable to resolve either
            # global *OR* local referees that are defined dynamically. Ergo, we
            # have no recourse but to detect this edge case and raise a
            # human-readable exception advising the caller with recommendations.
            if hint_ref_name == hint_ref_name_prev:
                raise BeartypeDecorHintPep695Exception(
                    f'{exception_prefix}PEP 695 local type alias "{hint_name}" '
                    f'unquoted relative forward reference "{hint_ref_name}" '
                    f"unsupported, due to severe deficiencies in CPython's "
                    f'runtime implementation of PEP 695 local type aliases '
                    f"outside beartype's control. Consider either:\n"
                    f'* Refactoring this local type alias into a '
                    f'global type alias:\n'
                    f'      # Instead of a local type alias '
                    f'defined in a callable like this...\n'
                    f'      def muh_func(...) -> ...:\n'
                    f'          type {hint_name} = ...\n'
                    f'\n'
                    f'      # Prefer a global type alias defined at module scope.\n'
                    f'      type {hint_name} = ...\n'
                    f'* Quoting this forward reference in this type alias:\n'
                    f'      # Instead of an unquoted forward reference '
                    f'like this...\n'
                    f'      type {hint_name} = ... {hint_ref_name} ...\n'
                    f'\n'
                    f'      # Prefer a quoted forward reference.\n'
                    f'      type {hint_name} = ... "{hint_ref_name}" ...'
                ) from exception
            # Else, this attribute differs from that of the prior iteration of
            # this "while" loop.
            #
            # If that module paradoxically claims to already define this
            # attribute as a global variable, raise an exception.
            #
            # Note that this should *NEVER* happen. Of course, this will happen.
            elif hasattr(hint_module, hint_ref_name):
                raise BeartypeDecorHintPep695Exception(  # pragma: no cover
                    f'{exception_prefix}PEP 695 type alias "{hint_name}" '
                    f'unquoted relative forward reference "{hint_ref_name}" '
                    f'already defined in module "{hint_module_name}", '
                    f'despite purportedly being undefined. '
                    f'In theory, this should never happen. '
                    f'Of course, this happened. You suddenly feel the '
                    f'horrifying urge to report this grievous failure to the '
                    f'beartype issue tracker:\n\t{URL_ISSUES}'
                ) from exception
            # Else, that module does *NOT* yet define this attribute.

            # Forward reference proxy to this undeclared attribute.
            #
            # Note that:
            # * This call is intentionally passed only positional parameters to
            #   satisfy the @callable_cached decorator memoizing this function.
            # * A full-blown forward reference proxy rather than a trivial
            #   stringified forward reference (i.e., the relative name of the
            #   undefined attribute being referred to, equivalent to
            #   "hint_ref_name") is required here. Why? Subscription. A
            #   stringified forward reference *CANNOT* be subscripted by
            #   arbitrary child type hints; a forward reference proxy can be.
            hint_ref = make_forwardref_indexable_subtype(
                hint_module_name, hint_ref_name)

            # Yield this forward reference proxy to the caller.
            yield hint_ref

            # Store the unqualified basename of this previously undeclared
            # attribute for detection by the next iteration of this loop.
            hint_ref_name_prev = hint_ref_name



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **import hook decorator-hostile intraprocess unit tests** (i.e.,
exercising edge cases of :mod:`beartype.claw` import hooks unique to
**decorator-hostile decorators** (i.e., decorators hostile to other decorators
by prohibiting other decorators from being applied after they are applied in a
chain of one or more decorators) within the active Python process).
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: Isolate each unit test defined below to its own subprocess. Why?
# Module imports. Since each unit test defined below tends to reimport the same
# (or, at least, similar) modules as previously run unit tests defined below,
# module imports and thus unit tests *MUST* be isolated to their own
# subprocesses to ensure these tests may be run in any arbitrary order.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
import pytest

# ....................{ TESTS                              }....................
@pytest.mark.run_in_subprocess
def test_claw_intraprocess_decorator_hostile() -> None:
    '''
    Test the :mod:`beartype.claw.beartype_package` import hook against a single
    data subpackage in this test suite exercising *all* edge cases associated
    with this import hook unique to **decorator-hostile decorators** (i.e.,
    decorators hostile to other decorators by prohibiting other decorators from
    being applied after they are applied in a chain of one or more decorators).
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.claw import beartype_package
    from beartype.claw._clawstate import (
        claw_lock,
        claw_state,
    )
    from beartype.roar import BeartypeDecorWrappeeException
    from beartype._conf.decorplace.confplacetrie import (
        BeartypeDecorPlacePackagesTrie,
        BeartypeDecorPlacePackageTrie,
        BeartypeDecorPlaceTypeTrie,
    )
    from beartype._data.conf import dataconfplace
    from beartype_test.a00_unit.data.func.data_decor import decorator_hostile
    from pytest import raises

    # ....................{ MONKEY-PATCH                   }....................
    #FIXME: *TRASH.* This is crude, unsafe, and frankly dumb. Instead, we should
    #just define a new "BeartypeConf" recognizing this testing-specific
    #decorator-hostile decorator as such. Sadly, "BeartypeConf" currently fails
    #to provide an option enabling this. *sigh*

    # Monkey-patch these testing-specific decorator-hostile decorator attributes
    # into this decorator-hostile decorator attribute name trie. Note that:
    # * The default contents of this trie describe decorator-hostile decorators
    #   only defined by third-parties and thus unsuitable for general-purpose
    #   unit testing.
    # * This monkey-patch is intentionally isolated to this test-specific
    #   subprocess and thus implicitly safe.
    dataconfplace.DECOR_HOSTILE_ATTR_NAME_TRIE |= (
        BeartypeDecorPlacePackagesTrie({
            'beartype_test': BeartypeDecorPlacePackageTrie({
                'a00_unit': BeartypeDecorPlacePackageTrie({
                    'data': BeartypeDecorPlacePackageTrie({
                        'func': BeartypeDecorPlacePackageTrie({
                            'data_decor': BeartypeDecorPlacePackageTrie({
                                'decorator_hostile': None,
                                'DecoratorHostileType': (
                                    BeartypeDecorPlaceTypeTrie({
                                        'decorator_hostile_method': None,
                                    })
                                ),
                            })
                        })
                    })
                })
            })
        })
    )
    # print(dataconfplace.DECOR_HOSTILE_ATTR_NAME_TRIE)

    # With a submodule-specific thread-safe reentrant lock, reset our import
    # hook state back to its initial defaults to respect the above monkey-patch.
    with claw_lock:
        claw_state.reinit()

    # ....................{ PREAMBLE                       }....................
    # Validate that the decorator-hostile decorator leveraged by the package
    # hooked below is a decorator-hostile decorator *BEFORE* importing from that
    # package. Since the @beartype decorator is permissive by both design and
    # necessity, defining decorator-hostile decorators unsupported by the
    # @beartype decorator is surprisingly non-trivial. Prove we actually did so.
    with raises(BeartypeDecorWrappeeException):
        @beartype
        @decorator_hostile
        def even_now_while_Saturn() -> None:
            '''
            Arbitrary callable decorated by a decorator-hostile decorator.
            '''

            pass

    # ....................{ LOCALS                         }....................
    # Name of the single package defining submodules defining callables and
    # types decorated by one or more decorator-hostile decorators to be subject
    # to beartype import hooks below.
    PACKAGE_NAME = (
        'beartype_test.a00_unit.data.claw.intraprocess.hookable_package.decor_hostile')

    # ....................{ PASS                           }....................
    # Explicitly subject this single package to a beartype import hook
    # configured by the default beartype configuration.
    beartype_package(PACKAGE_NAME)

    #FIXME: Uncomment after this actually works, please. *sigh*
    # Import the package hooked above, which then imports all submodules of that
    # package, exercising that these submodules are transitively subject to that
    # import hook.
    from beartype_test.a00_unit.data.claw.intraprocess.hookable_package import (
        decor_hostile)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484`- and :pep:`585`-compliant **subclass hint reducers** (i.e.,
low-level callables converting higher-level subscripted and unsubscripted
:pep:`484`-compliant ``typing.Type[...]`` and :pep:`585`-compliant ``type[...]``
hints to lower-level hints more readily consumable by :mod:`beartype`).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._check.metadata.hint.hintsane import HINT_SANE_IGNORABLE
from beartype._data.typing.datatypingport import Hint
from beartype._util.hint.pep.proposal.pep484585.pep484585 import (
    get_hint_pep484585_arg)

# PEP 484-specific type hint factories intentionally imported as such.
from typing import (
    Type as typing_Type,
)

# ....................{ REDUCERS                           }....................
def reduce_hint_pep484585_type(
    hint: Hint, exception_prefix: str, **kwargs) -> Hint:
    '''
    Reduce the passed :pep:`484`- or :pep:`585`-compliant **subclass hint**
    (i.e., hint constraining objects to subclass that superclass) to the
    :class:`type` superclass if that hint is subscripted by an ignorable child
    hint *or* preserve this hint as is otherwise (i.e., if this hint is *not*
    subscripted by an ignorable child hint).

    This reducer is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as reducers cannot be memoized.

    Parameters
    ----------
    hint : Hint
        Subclass type hint to be reduced.
    exception_prefix : str
        Human-readable label prefixing the representation of this object in the
        exception message.

    All remaining passed keyword parameters are silently ignored.

    Returns
    -------
    Hint
        Lower-level hint reduced from this subclass hint.

    Raises
    ------
    BeartypeDecorHintPep484585Exception
        If this hint is neither a :pep:`484`- nor :pep:`585`-compliant subclass
        hint.
    '''

    # Avoid circular import dependencies.
    from beartype._check.convert._reduce._pep.pep484.redpep484 import (
        reduce_hint_pep484_deprecated)
    from beartype._check.convert._reduce.redmain import reduce_hint_child

    # If this is a PEP 484-compliant subclass hint, this hint has been
    # deprecated by PEP 585. In this case, issue a non-fatal warning.
    reduce_hint_pep484_deprecated(hint=hint, exception_prefix=exception_prefix)

    # If this hint is the unsubscripted PEP 484-compliant subclass type hint,
    # immediately reduce this hint to the "type" superclass.
    #
    # Note that this is *NOT* merely a nonsensical optimization. The
    # implementation of the unsubscripted PEP 484-compliant subclass type hint
    # significantly differs across Python versions. Under some but *NOT* all
    # supported Python versions (notably, Python 3.7 and 3.8), the "typing"
    # module subversively subscripts this hint by a type variable; under all
    # others, this hint remains unsubscripted. In the latter case, passing this
    # hint to the subsequent get_hint_pep484585_args() call would erroneously
    # raise an exception.
    if hint is typing_Type:
        # print(f'Reducing subclass hint {hint} to "type"...')
        hint = type  # pyright: ignore
    # Else, this hint is *NOT* the unsubscripted PEP 484-compliant subclass
    # type hint. In this case...
    else:
        # Superclass subscripting this hint.
        #
        # Note that we intentionally do *NOT* call the high-level
        # get_hint_pep484585_type_superclass() getter here, as the
        # validation performed by that function would raise exceptions for
        # various child type hints that are otherwise permissible (e.g.,
        # "typing.Any").
        hint_child = get_hint_pep484585_arg(
            hint=hint, exception_prefix=exception_prefix)

        # Lower-level child hint reduced from this higher-level child hint.
        hint_child_reduced = reduce_hint_child(hint_child, kwargs)

        # If this child hint is ignorable, reduce this subclass hint to merely
        # the "type" superclass.
        if hint_child_reduced is HINT_SANE_IGNORABLE:
            # print(f'Reducing subclass hint {hint} to "type"...')
            hint = type  # pyright: ignore
        # Else, this child hint is unignorable. Preserve this hint as is.

        #FIXME: [SPEED] Consider uncommenting this optimization at a later date.
        #Doing so is complicated by the fact that it's currently insufficient;
        #we'd also need to consider the case in which "hint_child_reduced" is
        #metadata encapsulating a child hint reduction. *sigh*
        # # If this child hint was reduced to a different hint, preserve this
        # # reduction by re-subscripting this type hint factory by this reduction.
        # elif hint_child_reduced is not hint_child:
        #     hint = type[hint_child_reduced]
        # # Else, this child hint is irreducible. In this case, preserve this
        # # hint as is.

    # Return this possibly reduced hint.
    return hint



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype Inferential Type-hint Engine (BITE) third-party type hint inferrers**
(i.e., lower-level functions dynamically inferring subscripted type hints
describing non-standard popular third-party objects such as NumPy arrays and
PyTorch tensors).
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import (
    Annotated,
    Callable,
    Dict,
    Optional,
)
from beartype._util.api.external.utilnumpy import reduce_numpy_dtype
from beartype._util.module.utilmodget import get_object_module_name_or_none

# ....................{ INFERERS                           }....................
#FIXME: Unit test us up, please.
def infer_hint_thirdparty(obj: object, **kwargs) -> Optional[object]:
    '''
    **Third-party type hint** (i.e., possibly subscripted type hint factory
    published by a non-standard popular third-party package) recursively
    validating the passed object (including *all* items transitively reachable
    from this object) if this object derives from such a package *or*
    :data:`None` otherwise (i.e., if this object derives from *no* such
    package).

    This function *cannot* be memoized, due to necessarily accepting the
    ``__beartype_obj_ids_seen__`` parameter unique to each call to the parent
    :func:`beartype.bite.infer_hint` function. Moreover, this function exhibits
    worst-case constant time complexity :math:`O(1)`; memoization is irrelevant.

    Parameters
    ----------
    obj : object
        Object to infer a type hint from.

    All remaining keyword parameters are silently ignored.

    Returns
    -------
    Optional[object]
        Either:

        * If this object derives from a non-standard popular third-party
          package, a possibly subscripted type hint factory published by that
          package validating this object.
        * Else, :data:`None`.
    '''

    # Type of this object.
    obj_type = obj.__class__

    # Hint to be returned, defaulting to "None" as a fallback.
    hint: object = None

    # Fully-qualified name of the package declaring this type if any *OR* "None"
    # otherwise (e.g., if this type is dynamically declared in-memory and thus
    # resides outside any package structure).
    obj_type_package_name = get_object_module_name_or_none(obj_type)

    # If a package declares this type...
    if obj_type_package_name:
        # Dictionary mapping from the unqualified basename of that type to the
        # type hint inferer inferring type hints for this type if this is a
        # non-standard popular third-party type supported by this submodule *OR*
        # "None" otherwise (i.e., if this type is unsupported).
        type_basename_to_inferer = (
            _PACKAGE_NAME_TO_TYPE_BASENAME_TO_INFERER_get(
                obj_type_package_name))

        # If this package is recognized...
        if type_basename_to_inferer is not None:
            # Type hint inferer inferring type hints for this type if this is a
            # non-standard popular third-party type supported by this submodule
            # *OR* "None" otherwise (i.e., if this type is unsupported).
            #
            # Note that, unlike the "__module__" dunder attribute, the
            # "__name__" is effectively required to both exist and be non-empty
            # for all possible types -- even types declared in-memory.
            hint_inferer = type_basename_to_inferer.get(obj_type.__name__)

            # If this type is supported...
            if hint_inferer is not None:
                # Hint inferred by this type hint inferer.
                hint = hint_inferer(obj, **kwargs)
            # Else, this type is unsupported. In this case, fallback to
            # returning "None".
        # Else, this package is recognized. In this case, fallback to returning
        # "None".
    # Else, *NO* package declares this object. In this case, fallback to
    # returning "None".

    # Return this hint.
    return hint

# ....................{ PRIVATE ~ inferers                 }....................
def _infer_hint_thirdparty_numpy_ndarray(obj: object, **kwargs) -> object:
    '''
    **NumPy array type hint** validating the passed NumPy array.

    This function creates and returns a new :pep:`593`-compliant
    :obj:`typing.Annotated` type hint subscripted by (in order):

    * A non-standard :obj:`numpy.typing.NDArray` type hint factory subscripted
      by the **dtype** of the passed NumPy array.
    * A beartype-specific :obj:`beartype.vale.IsAttr` validator validating the
      **dimensionality** (i.e., :attr:`numpy.ndarray.ndim` instance variable) to
      be that of the passed NumPy array.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer package-specific imports.
    from beartype.vale import (
        IsAttr,
        IsEqual,
    )
    from numpy import ndarray  # pyright: ignore
    from numpy.typing import NDArray  # type: ignore[attr-defined]

    # Validate sanity.
    assert isinstance(obj, ndarray), f'{repr(obj)} not NumPy array.'

    # ....................{ DTYPE                          }....................
    # Coarse-grained NumPy-agnostic builtin type (e.g., "int") reduced from the
    # fine-grained NumPy-specific data type of the passed NumPy array (e.g.,
    # "numpy.int64"). Everybody wants the former. Nobody wants the latter.
    numpy_dtype = reduce_numpy_dtype(obj.dtype)  # type: ignore[name-defined]

    # Hint to be returned, defaulting to the "NDArray" type hint factory
    # subscripted by the dtype .
    hint: object = NDArray[numpy_dtype]  # type: ignore[misc,valid-type]

    # ....................{ SHAPE                          }....................
    # Generalize this hint to additionally validate the dimensionality
    # (i.e., number of dimensions equivalent to the length of the
    # "obj.shape" property) to be that of this array.
    hint = Annotated[hint, IsAttr['ndim', IsEqual[obj.ndim]]]  # type: ignore[name-defined]
    # Else, "typing(|_extensions).Annotated" is unimportable. In this case...

    # ....................{ RETURN                         }....................
    # Return this hint.
    return hint

# ....................{ PRIVATE ~ globals                  }....................
_PACKAGE_NAME_TO_TYPE_BASENAME_TO_INFERER: Dict[str, Dict[str, Callable]] = {
    # NumPy.
    'numpy': {
        # NumPy array.
        'ndarray': _infer_hint_thirdparty_numpy_ndarray,
    }
}
'''
Dictionary mapping from the fully-qualified package name of each non-standard
popular third-party type supported by this submodule to a nested dictionary
mapping from the unqualified basename of that type to the type hint inferer
inferring type hints for this type.

This data structure enables efficient constant time :math:`O(1)` lookups of
third-party types *without* requiring inefficient string munging.
'''


_PACKAGE_NAME_TO_TYPE_BASENAME_TO_INFERER_get = (
    _PACKAGE_NAME_TO_TYPE_BASENAME_TO_INFERER.get)
'''
:meth:`dict.get` method bound to the global
:data:`_PACKAGE_NAME_TO_TYPE_BASENAME_TO_INFERER` dictionary for negligible
lookup efficiency.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **Python expression indentation substring unit tests.**

This submodule unit tests the public API of the public
:mod:`beartype._data.hint.sign.datahintsignset` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                              }....................
def test_data_indent_level_to_code() -> None:
    '''
    Test the :obj:`beartype._data.code.datacodeindent.INDENT_LEVEL_TO_CODE`
    dictionary singleton.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._data.code.datacodeindent import INDENT_LEVEL_TO_CODE
    from pytest import raises

    # ....................{ PASS                           }....................
    # Assert this dictionary indexed by various indentation levels creates and
    # returns the expected indentation string constants.
    assert INDENT_LEVEL_TO_CODE[1] == '    '
    assert INDENT_LEVEL_TO_CODE[2] == '        '
    assert INDENT_LEVEL_TO_CODE[3] == '            '

    # Assert this dictionary internally caches these constants.
    assert INDENT_LEVEL_TO_CODE[1] is INDENT_LEVEL_TO_CODE[1]
    assert INDENT_LEVEL_TO_CODE[2] is INDENT_LEVEL_TO_CODE[2]
    assert INDENT_LEVEL_TO_CODE[3] is INDENT_LEVEL_TO_CODE[3]

    # ....................{ FAIL                           }....................
    # Assert that attempting to index this dictionary by non-integer
    # indentation levels raises the expected exception.
    with raises(AssertionError):
        INDENT_LEVEL_TO_CODE[2.34]

    # Assert that attempting to index this dictionary by non-positive
    # indentation levels raises the expected exception.
    with raises(AssertionError):
        INDENT_LEVEL_TO_CODE[0]
    with raises(AssertionError):
        INDENT_LEVEL_TO_CODE[-1]



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype decorator :pep:`563` unit tests.

This submodule unit tests :pep:`563` support implemented in the
:func:`beartype.beartype` decorator.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
# WARNING: The "from __future__ import annotations" import appears to be subtly
# broken under Python >= 3.10, where performing that import prevents the
# "f_locals" attribute of stack frames from capturing locals defined by parent
# callables and accessed *ONLY* in type hints annotating one or more parameters
# of nested callables contained in those parent callables. Oddly, this breakage
# does *NOT* extend to type hints annotating returns of those nested callables
# -- only parameters. Ergo, we can only conclude this to be a subtle bug. We
# should probably issue an upstream CPython report. For now, conditionally
# disable *ALL* PEP 563-specific tests importing from another module containing
# an "from __future__ import annotations" import.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

from beartype_test._util.mark.pytskip import (
    skip_if_pypy,
    skip_if_python_version_less_than,
    skip_if_python_version_greater_than_or_equal_to,
)

# .....................{ TESTS ~ club                      }....................
def test_pep563_class_self_reference_reloaded() -> None:
    '''
    Test module-scoped :pep:`563` support implemented in the
    :func:`beartype.beartype` decorator.

    This test exercises multiple issues, including:

    * `An issue <issue #49_>`__ concerning a :pep:`563`-postponed circular
      reference to a class currently being declared from a method of that class
      decorated by :func:`beartype.beartype`. See the
      :meth:`beartype_test.a00_unit.data.pep.pep563.data_pep563_club.Chameleon.like_a_dream`
      docstring for further details.
    * `An issue <issue #152_>`__ concerning a :pep:`563`-postponed circular
      reference to a class currently being declared embedded in a parent type
      hint of a method of that class decorated by :func:`beartype.beartype`. See
      the
      :meth:`beartype_test.a00_unit.data.pep.pep563.data_pep563_club.Chameleon.when_we_cling`
      docstring for further details.

    .. _issue #49:
       https://github.com/beartype/beartype/issues/49
    .. _issue #152:
       https://github.com/beartype/beartype/issues/152
    '''

    # .....................{ IMPORTS                       }....................
    # Defer test-specific imports.
    #
    # Note that the "data_pep563_club" submodule is intentionally imported as an
    # attribute rather than importing the requisite attributes from that
    # submodule. Why? Because the entire intention of this test is to exercise
    # reloading of @beartype-decorated callables annotated with circular
    # references under PEP 563.
    from beartype_test.a00_unit.data.pep.pep563 import data_pep563_club
    from importlib import reload

    # .....................{ PASS                          }....................
    # Assert that a @beartype-decorated class method whose circular return
    # annotation is postponed under PEP 563 returns the expected value.
    assert data_pep563_club.Chameleon.like_my_dreams().colors == (
        data_pep563_club.COLORS)

    # Assert that a @beartype-decorated class method whose circular return
    # annotation is embedded in a parent type hint postponed under PEP 563
    # returns the expected value.
    assert data_pep563_club.Chameleon.when_we_cling().colors == (
        data_pep563_club.CLING)

    # Intentionally reload this submodule into the same attribute.
    print('Reloading submodule "data_pep563_club"...')
    data_pep563_club = reload(data_pep563_club)
    print('Reloaded submodule "data_pep563_club".')

    # Re-assert that a @beartype-decorated class method whose circular return
    # annotation is postponed under PEP 563 returns the expected value.
    #
    # Note that this method continues to behave as expected despite the
    # submodule defining this class (and thus this method) being reloaded above.
    # Why? Because the return annotating this method actually reduces to a
    # relative forward reference dynamically resolved at method call time.
    assert data_pep563_club.Chameleon.like_my_dreams().colors == (
        data_pep563_club.COLORS)

    #FIXME: This assertion currently fails, which is almost certainly a bug
    #pertaining to internal caching of code generated by @beartype... maybe?
    #importlib.reload() is well-known to have harmful interactions with various
    #low-level Python machinery. Perhaps this is simply one of them. Let's leave
    #this unresolved until someone angrily complains about this. \o/
    # assert data_pep563_club.Chameleon.when_we_cling().colors == (
    #     data_pep563_club.CLING)


def test_pep563_class_self_reference_override() -> None:
    '''
    Test module-scoped :pep:`563` support implemented in the
    :func:`beartype.beartype` decorator.

    This test exercises an edge case in which a :pep:`563`-postponed circular
    reference to a class currently being declared from a method of that class
    decorated by :func:`beartype.beartype` overrides and thus conflicts with a
    previously declared global with the same name as that class in the module
    declaring that class.
    '''

    # Defer test-specific imports.
    from beartype_test.a00_unit.data.pep.pep563.data_pep563_club import (
        DREAMS, Karma)

    # Assert that a @beartype-decorated class method whose circular return
    # annotation is postponed under PEP 563 to a string of the same name as a
    # previously declared global of the module declaring that class returns the
    # expected value.
    assert Karma.if_your_colors().dreams == DREAMS

# ....................{ TESTS ~ scope                      }....................
def test_pep563_module() -> None:
    '''
    Test module-scoped :pep:`563` support implemented in the
    :func:`beartype.beartype` decorator.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype._util.hint.pep.proposal.pep649 import (
        get_pep649_hintable_annotations,
        set_pep649_hintable_annotations,
    )
    from beartype_test.a00_unit.data.pep.pep563.data_pep563_poem import (
        get_minecraft_end_txt,
        get_minecraft_end_txt_pep604,
        get_minecraft_end_txt_stanza,
    )

    # ....................{ LOCALS                         }....................
    # Dictionary of these callables' annotations, localized to enable debugging
    # in the likely event of unit test failure. *sigh*
    get_minecraft_end_txt_annotations = get_pep649_hintable_annotations(
        get_minecraft_end_txt)
    get_minecraft_end_txt_stanza_annotations = get_pep649_hintable_annotations(
        get_minecraft_end_txt_stanza)

    # from annotationlib import Format
    # hintable = get_minecraft_end_txt_stanza
    # print(f'{hintable}.__annotate__: {hintable.__annotate__}')
    # print(f'{hintable}.__annotations__: {hintable.__annotations__}')
    # print(f'{hintable}.__annotate__(3): {hintable.__annotate__(Format.FORWARDREF)}')
    # hintable_annotations_cached = get_pep649_hintable_annotations(hintable)
    # print(f'{hintable}.__annotate__(3) [cached]: {hintable_annotations_cached}')

    # ....................{ ASSERTS                        }....................
    # Assert that all annotations of a callable *NOT* decorated by @beartype
    # are postponed under PEP 563 as expected.
    assert all(
        isinstance(param_hint, str)
        for arg_name, param_hint in get_minecraft_end_txt_annotations.items()
    )

    # Assert that *NO* annotations of a @beartype-decorated callable are
    # postponed, as @beartype implicitly resolves all annotations.
    assert all(
        not isinstance(param_hint, str)
        for arg_name, param_hint in (
            get_minecraft_end_txt_stanza_annotations.items())
    )

    # Assert that a @beartype-decorated callable works under PEP 563.
    assert get_minecraft_end_txt_stanza(
        player_name='Notch', stanza_index=33) == 'Notch. Player of games.'

    # Test that @beartype silently accepts callables with one or more
    # non-postponed annotations under PEP 563, a technically non-erroneous edge
    # case that needlessly complicates code life.
    #
    # Manually resolve all postponed annotations on a callable.
    set_pep649_hintable_annotations(
        hintable=get_minecraft_end_txt,
        annotations={
            arg_name: eval(param_hint, get_minecraft_end_txt.__globals__)
            for arg_name, param_hint in (
                get_pep649_hintable_annotations(get_minecraft_end_txt).items())
        },
    )

    # ....................{ ASSERTS ~ decorator            }....................
    # Manually decorate this callable with @beartype.
    get_minecraft_end_txt_typed = beartype(get_minecraft_end_txt)

    # Manually decorate a PEP 604-compliant callable with @beartype.
    get_minecraft_end_txt_pep604_typed = beartype(get_minecraft_end_txt_pep604)

    # Assert that these callables behave as expected under PEP 563.
    assert isinstance(
        get_minecraft_end_txt_typed(player_name='Notch'), str)
    assert isinstance(
        get_minecraft_end_txt_pep604_typed(player_name='Notch'), str)


def test_pep563_class() -> None:
    '''
    Test class-scoped :pep:`563` support implemented in the
    :func:`beartype.beartype` decorator.
    '''

    # Defer test-specific imports.
    from beartype_test.a00_unit.data.pep.pep563.data_pep563_poem import (
        MinecraftEndTxtUnscrambler)

    # Assert that instantiating a class with a @beartype-decorated __init__()
    # method declaring a @beartype-decorated method closure works.
    minecraft_end_txt_unscrambler = MinecraftEndTxtUnscrambler(
        unscrambling='dream')

    # Assert that that __init__() method declared that method closure.
    get_minecraft_end_txt_unscrambled_stanza = (
        minecraft_end_txt_unscrambler.get_minecraft_end_txt_unscrambled_stanza)
    assert callable(get_minecraft_end_txt_unscrambled_stanza)

    # Assert that this method closure works under PEP 563.
    minecraft_end_txt_unscrambled_stanza = (
        get_minecraft_end_txt_unscrambled_stanza(
            minecraft_end_txt_unscrambler, is_stanza_last=True))
    assert isinstance(minecraft_end_txt_unscrambled_stanza, str)
    assert minecraft_end_txt_unscrambled_stanza.count('dream') == 5


def test_pep563_closure_nonnested() -> None:
    '''
    Test non-nested closure-scoped :pep:`563` support implemented in the
    :func:`beartype.beartype` decorator.
    '''

    # Defer test-specific imports.
    from beartype_test.a00_unit.data.pep.pep563.data_pep563_poem import (
        get_minecraft_end_txt_closure)

    # Assert that declaring a @beartype-decorated closure works under PEP 563.
    get_minecraft_end_txt_substr = get_minecraft_end_txt_closure(
        player_name='Julian Gough')
    assert callable(get_minecraft_end_txt_substr)

    # Assert that this closure works under PEP 563.
    minecraft_end_txt_substr = get_minecraft_end_txt_substr('player')
    assert isinstance(minecraft_end_txt_substr, list)
    assert 'You are the player.' in minecraft_end_txt_substr


#FIXME: *REPORT AN UPSTREAM ISSUE WITH THE PYPY TRACKER AT:*
#    https://foss.heptapod.net/pypy/pypy/-/issues
#
#So, what's going on here? What's going on here is that PyPy appears to be
#slightly broken with respect to the "FrameType.f_locals" dictionary. Whereas
#CPython provides the actual dictionary of lexically scoped locals required by
#the current frame, PyPy only dynamically computes that dictionary on-the-fly.
#This dynamic computation is known to fail cross-thread but was assumed to be
#fully compliant with CPython expectations when running in the same thread.
#
#This does *NOT* appear to be the case. Specifically, the "IntLike" local
#variable declared by the top-level get_minecraft_end_txt_closure_factory()
#function is:
#* Correctly accessible to the lowest-level
#  get_minecraft_end_txt_closure_inner() closure with respect to actual
#  interpretation of statements and thus declaration of closures by PyPy.
#* Incorrectly omitted from the "FrameType.f_locals" dictionary of the frame
#  object for the mid-level get_minecraft_end_txt_closure_outer() closure
#  declaring the lowest-level get_minecraft_end_txt_closure_inner() closure.
#  What's bizarre is that that dictionary correctly includes the "player_name"
#  parameter accepted by the top-level get_minecraft_end_txt_closure_factory()
#  function.
#
#Clearly, what's happening here is that PyPy developers failed to add local
#variables of lexical scopes declared by distant parent callables (i.e.,
#callables that are *NOT* the direct parent of the lowest-level closure in
#question) when those local variables are *ONLY* accessed in annotations.
#Moreover, when those local variables are accessed outside annotations (as
#with the "player_name" parameter) in the body of the lowest-level closure,
#those variables are correctly added to the "FrameType.f_locals" dictionary.
#Ergo, this is an annotation-specific issue in the internal algorithm PyPy
#uses to dynamically construct that dictionary on-the-fly. Admittedly, this was
#an edge case that basically didn't matter until PEP 563 landed -- at which
#point this edge case *REALLY* mattered.
#FIXME: CPython is subtly broken with respect to "from __future__ import
#annotations" imports under Python >= 3.10. Until resolved, disable this.

@skip_if_pypy()
@skip_if_python_version_greater_than_or_equal_to('3.10.0')
def test_pep563_closure_nested() -> None:
    '''
    Test nested closure-scoped :pep:`563` support implemented in the
    :func:`beartype.beartype` decorator.
    '''

    # Defer test-specific imports.
    from beartype_test.a00_unit.data.pep.pep563.data_pep563_poem import (
        get_minecraft_end_txt_closure_factory)

    # Assert that declaring a @beartype-decorated closure factory works under
    # PEP 563.
    get_minecraft_end_txt_closure_outer = (
        get_minecraft_end_txt_closure_factory(player_name='Markus Persson'))
    assert callable(get_minecraft_end_txt_closure_outer)

    # Assert that declaring a @beartype-decorated closure declared by a
    # @beartype-decorated closure factory works under PEP 563.
    get_minecraft_end_txt_closure_inner = get_minecraft_end_txt_closure_outer(
        stanza_len_min=65)
    assert callable(get_minecraft_end_txt_closure_inner)

    # Assert that this closure works under PEP 563.
    minecraft_end_txt_inner = get_minecraft_end_txt_closure_inner(
        stanza_len_max=65, substr='thought')
    assert isinstance(minecraft_end_txt_inner, list)
    assert len(minecraft_end_txt_inner) == 1
    assert minecraft_end_txt_inner[0] == (
        'It is reading our thoughts as though they were words on a screen.')

# .....................{ TESTS ~ pep : 484                 }....................
def test_pep563_hint_pep484_namedtuple() -> None:
    '''
    Test module-scoped :pep:`563` support implemented in the
    :func:`beartype.beartype` decorator with respect to :pep:`484`-compliant
    :obj:`typing.NamedTuple` subclasses decorated by this decorator.
    '''

    # .....................{ IMPORTS                       }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeCallHintParamViolation
    from beartype_test.a00_unit.data.pep.pep563.pep484.data_pep563_pep484 import (
        LeadOnlyTo)
    from pytest import raises

    # .....................{ PASS                          }....................
    # Assert that instantiating this subclass with valid fields creates and
    # returns an object of the expected type exposing the same fields.
    while_deaths_blue_vault = LeadOnlyTo(0xCAFEFACE)
    assert while_deaths_blue_vault.a_black_and_watery_depth == 0xCAFEFACE

    # .....................{ FAIL                          }....................
    # Assert that instantiating this subclass with one or more invalid fields
    # raises the expected exception.
    with raises(BeartypeCallHintParamViolation):
        LeadOnlyTo("While death's blue vault, with loathliest vapours hung,")


def test_pep563_hint_pep484_noreturn() -> None:
    '''
    Test module-scoped :pep:`563` support implemented in the
    :func:`beartype.beartype` decorator with respect to callables with returns
    annotated by the :pep:`484`-compliant :obj:`typing.NoReturn` type hint.
    '''

    # Defer test-specific imports.
    from beartype_test.a00_unit.data.pep.pep563.data_pep563_club import (
        HeWouldLingerLong,
        to_love_and_wonder,
    )
    from pytest import raises

    # Assert that a @beartype-decorated function "typing.NoReturn" return
    # annotation is postponed under PEP 563 to a string raises the expected
    # exception.
    with raises(HeWouldLingerLong):
        to_love_and_wonder()

# .....................{ TESTS ~ pep : 604                 }....................
@skip_if_python_version_less_than('3.10.0')
def test_pep563_hint_pep604() -> None:
    '''
    Test module-scoped :pep:`563` support implemented in the
    :func:`beartype.beartype` decorator with respect to :pep:`604`-compliant
    unions annotating :pep:`557`-compliant
    :obj:`dataclasses.dataclass`-decorated subclasses also decorated by this
    decorator if the active Python interpreter targets Python >= 3.10 and thus
    supports :pep:`604`.
    '''

    # .....................{ IMPORTS                       }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeCallHintParamViolation
    from beartype_test.a00_unit.data.pep.pep563.pep604.data_pep563_pep604 import (
        FrameMoreAttuned,
        WithVoiceFarSweeter,
    )
    from pytest import raises

    # .....................{ LOCALS                        }....................
    # Arbitrary instance of a problematic dataclass subclass annotated by a
    # problematic PEP 604-compliant union that had yet to be defined at the time
    # that subclass was defined.
    wasting_these_surpassing_powers = WithVoiceFarSweeter(
        than_thy_dying_notes=FrameMoreAttuned())
    # wasting_these_surpassing_powers = WithVoiceFarSweeter()

    # .....................{ PASS                          }....................
    # Assert that this instance exposes the expected fields.
    isinstance(
        wasting_these_surpassing_powers.than_thy_dying_notes, FrameMoreAttuned)

    # .....................{ FAIL                          }....................
    # Assert that instantiating this subclass with one or more invalid fields
    # raises the expected exception.
    with raises(BeartypeCallHintParamViolation):
        WithVoiceFarSweeter(than_thy_dying_notes=(
            'In the deaf air, to the blind earth, and heaven'))

# .....................{ TESTS ~ pep : 695                 }....................
@skip_if_python_version_less_than('3.12.0')
def test_pep563_hint_pep695() -> None:
    '''
    Test module-scoped :pep:`563` support implemented in the
    :func:`beartype.beartype` decorator with respect to :pep:`695`-compliant
    unions annotating :pep:`557`-compliant
    :obj:`dataclasses.dataclass`-decorated subclasses also decorated by this
    decorator if the active Python interpreter targets Python >= 3.12 and thus
    supports :pep:`695`.
    '''

    # .....................{ IMPORTS                       }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeCallHintParamViolation
    from beartype_test.a00_unit.data.pep.pep563.pep695.data_pep563_pep695 import (
        NakedAndBare,
        shook_horrid,
    )
    from pytest import raises

    # .....................{ LOCALS                        }....................
    # Instance of a class parametrized by PEP 695-compliant type parameters.
    who_had_power = NakedAndBare()

    # .....................{ PASS                          }....................
    # Assert that calling this PEP 695-parametrized callable with valid
    # parameters returns the expected value.
    assert shook_horrid(
        ['Shook', 'horrid', 'with', 'such', 'aspen-malady']) == 'Shook'
    assert shook_horrid([7, 2, 4, 0]) == 7

    # Assert that calling a PEP 695-parametrized method of this instance with
    # valid parameters returns the expected value.
    assert who_had_power.of_its_great_diadem(
        ['To', 'make', 'me', 'desolate?', 'whence', 'came', 'the', 'strength?']
    ) == 'To'

    # .....................{ FAIL                          }....................
    # Assert that calling these PEP 695-parametrized callables with invalid
    # parameters raise the expected exception.
    with raises(BeartypeCallHintParamViolation):
        shook_horrid('"O tender spouse of gold Hyperion,')
    with raises(BeartypeCallHintParamViolation):
        who_had_power.of_its_great_diadem(
            "How was it nurtur'd to such bursting forth,")

# ....................{ TESTS ~ limit                      }....................
#FIXME: Hilariously, we can't even unit test whether the
#beartype._decor._pep.pep563._die_if_hint_repr_exceeds_child_limit() function
#behaves as expected. See commentary in the
#"beartype_test.a00_unit.data.pep.pep563.data_pep563_poem" submodule for all the appalling details.

# @skip_if_python_version_less_than('3.7.0')
# def test_die_if_hint_repr_exceeds_child_limit() -> None:
#     '''
#     Test the private
#     :func:`beartype._decor._pep.pep563._die_if_hint_repr_exceeds_child_limit`
#     function if the active Python interpreter targets at least Python 3.7.0
#     (i.e., the first major Python version to support PEP 563) *or* skip
#     otherwise.
#     '''
#
#     # Defer test-specific imports.
#     from beartype import beartype
#     from beartype.roar import BeartypeDecorHintPepException
#     from beartype_test.a00_unit.data.pep.pep563.data_pep563_poem import player_was_love
#
#     # Assert @beartype raises the expected exception when decorating a
#     # callable violating this limit.
#     with raises(BeartypeDecorHintPepException):
#         beartype(player_was_love)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype Inferential Type-hint Engine (BITE) collections abstract base class
(ABC) type hint inferrers** (i.e., lower-level functions dynamically inferring
subscripted type hints describing instances of standard :mod:`collections.abc`
protocols).
'''

# ....................{ TODO                               }....................
#FIXME: *NON-IDEAL.* This is mostly great. "collections.abc.Mapping" support is
#the exception, sadly; that's demonstrably *NOT* great. Why? Because the
#infer_hint_collections_abc() function currently lies. How so? By erroneously
#inferring that user-defined classes virtually satisfying the
#"collections.abc.Mapping" protocol *WITHOUT* physically subclassing either that
#protocol or the builtin "dict" type, which is registered with
#"collections.abc.Mapping" and thus a hardcoded "virtual subclass" of
#"collections.abc.Mapping" actually do satisfy "collections.abc.Mapping": e.g.,
#    >>> from beartype.bite import infer_hint
#    >>> from beartype_test.a00_unit.data.data_type import ClassMapping
#    >>> from collections.abc import Mapping
#
#    >>> map = ClassMapping({
#    ...     'Branchless and blasted,': b'clenched with grasping roots',
#    ...     'The unwilling soil.': b'A gradual change was here,',
#    ... })
#
#    >>> isinstance(map, Mapping)
#    False
#    >>> infer_hint(map)
#    Annotated[Mapping[str, bytes], IsInstance[ClassMapping]]
#
#That's totally wrong.
#
#On the one hand, the "collections.abc.Mapping" protocol erroneously asserts
#that "ClassMapping" instances do *NOT* satisfy that protocol. Why? Because,
#unlike *MOST* protocols, that protocol fails to define the __subclasshook__()
#dunder method required to support "virtual subclasses. "ClassMapping" instances
#do actually satisfy that protocol, of course, but we can't reasonably
#monkey-patch that awareness into the standard "collections.abc.Mapping" class
#without breaking everybody's downstream code.
#
#On the other other hand, our infer_hint() function infers that "ClassMapping"
#instances do actually satisfy that protocol. They don't, though, due to the
#aforementioned failure of that protocol to define the __subclasshook__()
#dunder method required to support "virtual subclasses."
#
#The solution? The genius-tier @JWCS solved this:
#* Improve our custom "beartype.typing.Mapping" ABC into a
#  "collections.abc.Mapping" subclass that additionally defines
#  __subclasshook__() in the standard way to support virtual subclasses.
#
#That's it. That's the entire solution. Ridiculous, huh? Genius is genius.

# ....................{ IMPORTS                            }....................
from beartype.typing import (
    TYPE_CHECKING,
    Annotated,
    Dict,
    Optional,
    Set,
)
from beartype._data.kind.datakindmap import FROZENDICT_EMPTY
from beartype._data.typing.datatyping import FrozenSetStrs
from beartype._util.cache.utilcachecall import callable_cached
from beartype._util.utilobjattr import get_object_methods_name_to_value_explicit
from collections.abc import (
    Collection as CollectionABC,
)

# ....................{ INFERERS                           }....................
def infer_hint_collections_abc(obj: object, **kwargs) -> Optional[object]:
    '''
    **Collections protocol type hint** (i.e., possibly subscripted abstract base
    class (ABC) published by the :mod:`collections.abc` subpackage) recursively
    validating the passed object (including *all* items transitively reachable
    from this object) if this object satisfies at least one such protocol *or*
    :data:`None` otherwise (i.e., if this object satisfies *no* such protocol).

    This function *cannot* be memoized, due to necessarily accepting the
    ``__beartype_obj_ids_seen__`` parameter unique to each call to the parent
    :func:`beartype.bite.infer_hint` function.

    Caveats
    -------
    This function exhibits:

    * **Best-case linear time complexity** :math:`O(n)` where :math:`n` is the
      number of attributes bound to the type of the passed object, which occurs
      when that class satisfies *no* :mod:`collections.abc` ABC.
    * **Average-case linear time complexity** :math:`O(n + k)` where :math:`n`
      is the number of attributes bound to the type of the passed object and
      :math:`k` is the number of :mod:`collections.abc` ABCs subclassing the
      :class:`collections.abc.Container` superclass, which occurs when that
      class is that of a **standard container** (i.e., satisfies at least the
      root :class:`collections.abc.Container` ABC).
    * **Worst-case quadratic time complexity** :math:`O(n + jk)` where :math:`n`
      is the number of attributes bound to the type of the passed object,
      :math:`j` is the number of :mod:`collections.abc` ABC superclasses (i.e.,
      ABCs at the root of a hierarchy of ABCs), and :math:`k` is the largest
      number of :mod:`collections.abc` ABCs subclassing such a superclass, which
      occurs when that class satisfies only an uncommon :mod:`collections.abc`
      ABC superclass.

    Parameters
    ----------
    obj : object
        Object to infer a type hint from.

    All remaining keyword parameters are passed as is to the lower-level
    :func:`beartype.bite.collection.infercollectionitems.infer_hint_collection_items`
    function.

    Returns
    -------
    Optional[object]
        Either:

        * If this object satisfies at least one :class:`collections.abc`
          protocol:
            * If this object is a **collection** (i.e., this ABC is a
              :class:`collections.abc.Collection` subclass), this ABC
              subscripted by the union of the child type hints validating *all*
              items of this collection.
            * Else, this ABC unsubscripted (i.e., as is) otherwise.
        * Else, :data:`None`.
    '''

    # Type of this object.
    obj_type = obj.__class__

    # Hint to be returned, defaulting to "None" as a fallback.
    hint: object = None

    # Narrowest "collections.abc" protocol validating this type if at least one
    # such protocol validates this type *OR* "None" otherwise.
    hint_factory = _infer_hint_factory_collections_abc(obj_type)

    # If at least one "collections.abc" protocol validates this type...
    if hint_factory:
        # Defer heavyweight imports.
        from beartype.vale import IsInstance

        # If this object is a collection...
        if isinstance(obj, CollectionABC):
            # Avoid circular import dependencies.
            from beartype.bite.collection.infercollectionitems import (
                infer_hint_collection_items)

            # "collections.abc" protocol validating this type, defined this hint
            # factory itself. Why? Because this hint factory is already its own
            # origin type due to being a valid "collections.abc" superclass.
            origin_type: type = hint_factory  # type: ignore[assignment]
            # print(f'Inferring iterable {repr(obj_type_collections_abc)} subscription...')

            # Hint recursively validating this iterable (including *ALL* items
            # transitively reachable from this iterable), defined by
            # subscripting this "collections.abc" protocol by the union of the
            # child type hints validating all items recursively reachable from
            # this iterable.
            hint = infer_hint_collection_items(
                obj=obj,  # type: ignore[arg-type]
                hint_factory=hint_factory,
                origin_type=origin_type,
                **kwargs
            )
        # Else, this protocol is *NOT* a container protocol. In this case, we
        # have *NO* safe means of inferring the child type hints subscripting
        # this protocol. In this case, fallback to returning this unsubscripted
        # container protocol as is.
        else:
            # print(f'Ignoring non-iterable {repr(obj_type_collections_abc)} subscription...')
            hint = hint_factory

        #FIXME: In fact, we can and should do even better. Notably:
        #* If the set of all *PUBLIC* attributes (i.e., attributes whose names
        #  are *NOT* prefixed by a single "_" character) bound to the class of
        #  the passed object is exactly identical to the set of all attributes
        #  required by this "collections.abc" protocol "origin_type", then this
        #  protocol actually does perfectly encapsulate this class. In this
        #  case, we should simply return this "hint" as is rather than a
        #  full-blown PEP 593-compliant "Annotated[...]" parent type hint
        #  encapsulating this "hint" as a child type hint.

        # Generalize this possibly subscripted "collections.abc" protocol
        # into a PEP 593-compliant "Annotated[...]" hint annotating this
        # protocol to additionally require that arbitrary objects satisfying
        # this hint also be instances of the same type as the passed object.
        #
        # Doing so avoids false positives, as the original "hint" (i.e.,
        # possibly subscripted "collections.abc" protocol) widely matches
        # *ANY* object of *ANY* type that simply satisfies that protocol.
        # However, the type of the passed object (probably) encapsulates
        # other attributes and behaviours *NOT* described by this protocol.
        # In and of itself, this protocol fails to fully validate other
        # objects of the same type.
        hint = Annotated[hint, IsInstance[obj_type]]
    # Else, *NO* "collections.abc" protocol validates this type. In this
    # case, fallback to returning "None".

    # Return this hint.
    return hint

# ....................{ PRIVATE ~ inferers                 }....................
@callable_cached
def _infer_hint_factory_collections_abc(cls: type) -> Optional[object]:
    '''
    Narrowest **collections protocol** (i.e., abstract base class published by
    the :mod:`collections.abc` subpackage) validating the passed class if at
    least one collections protocol validates this class *or* :data:`None`
    otherwise (i.e., if *no* collections protocol validates this class).

    This function returns the narrowest collections protocol, defined such that:

    * The passed class is a **virtual subclass** of (i.e., defines all methods
      required by) this protocol.
    * This protocol is **maximally narrow.** Formally:

      * The passed class is a virtual subclass of *no* other collections
        protocols that are themselves virtual subclasses of this protocol.
      * Equivalently, this protocol is **larger** (i.e., defines more abstract
        methods) than all other collections protocols that the passed class is
        also a virtual subclass of.

    This function does *not* return a type hint annotating the passed class.
    This function merely returns a subscriptable type hint factory suitable for
    the caller to subscript with a child type hint, which then produces the
    desired type hint annotating the passed class. Why this indirection? In a
    word: efficiency. If this function recursively inferred and returned a full
    type hint, this function would need to guard against infinite recursion by
    accepting the ``__beartype_obj_ids_seen__`` parameter also accepted by the
    parent :func:`beartype.bite.inferhint.infer_hint` function;
    doing so would prevent memoization, substantially reducing efficiency.

    This function is memoized for efficiency.

    Design
    ------
    This function efficiently infers the narrowest :mod:`collections.abc` ABC
    applicable to the passed object by introspecting the set of all **dunder
    methods** (i.e., methods whose names are both prefixed and suffixed by the
    double-underscore substring ``"__"``). To do so, this function internally
    performs a graph theoretic traversal over a finite state machine (FSM)
    whose nodes are the set of all :mod:`collections.abc` ABCs and transitions
    between those nodes the dunder method names required by those ABCs. For
    example, given a passed object defining only the standard
    ``__contains__()``, ``__iter__()``, and ``__len__()`` dunder methods, this
    function (in order):

    #. Transitions from the start node of this FSM along the ``__contains__()``
       edge of this FSM to the corresponding :class:`collections.abc.Container`
       node.
    #. Transitions from that node along the grouped ``{__iter__(), __len__()}``
       edge of this FSM to the corresponding :class:`collections.abc.Collection`
       node.
    #. Halts there, thus inferring the type hint factory validating the passed
       object to be the :class:`collections.abc.Collection` ABC.

    Parameters
    ----------
    cls : type
        Class to be introspected.

    Returns
    -------
    Optional[object]
        Either:

        * If at least one :mod:`collections.abc` protocol validates this class,
          the narrowest such protocol.
        * Else, :data:`None`.
    '''
    assert isinstance(cls, type), f'{repr(cls)} not type.'

    # ....................{ PREAMBLE                       }....................
    # List of the names of all attributes bound to this class.
    cls_attr_names = dir(cls)

    # "collections.abc" finite state machine (FSM).
    _START_NODE = get_finite_state_machine()

    # If the intersection of the set of the names of all attributes bound to
    # this class with the set of the names of all requisite methods (i.e.,
    # methods required to transition from the start node to a transitional node)
    # is the empty set, then this class fails to define *ANY* methods required
    # by a "collections.abc" ABC. In this case, silently reduce to a noop.
    if not (set(cls_attr_names) & _START_NODE.nodes_next_method_names):
        return None
    # Else, this intersection is non-empty. In this case, this class defines at
    # least one attribute whose name is that of a method required by a
    # "collections.abc" ABC.
    #
    # Note that this does *NOT* necessarily imply that this class defines at
    # least one such method. Clearly, an instance variable of the same name is
    # *NOT* a method: e.g.,
    #     # This class does *NOT* satisfy the "collections.abc.Sized" protocol
    #     # despite defining the "__sized__" attribute, as that attribute is
    #     # *NOT* a method.
    #     class FakeSized(object):
    #         __sized__: int
    #
    # Further detection is warranted to disambiguate this edge case.

    # ....................{ LOCALS                         }....................
    # Dictionary mapping from the names of all methods of the passed class to
    # those methods.
    cls_methods_name_to_method = get_object_methods_name_to_value_explicit(
        obj=cls, obj_dir=cls_attr_names)

    # Set of the names of all methods bound to this class.
    cls_method_names = cls_methods_name_to_method.keys()

    # Finite state machine (FSM) node currently visited by the search below.
    node_curr = _START_NODE

    # Finite state machine (FSM) node to be visited next by the search below if
    # any *OR* "None" if this is the last iteration of this search.
    node_next: Optional[_FiniteStateMachineNode] = None

    # ....................{ SEARCH                         }....................
    # print(f'Inferring {repr(cls)} "collections.abc" protocol...')

    # While some FSM node is still being visited by this search...
    while True:
        # print(f'Visiting FSM node {repr(node_curr.hint_factory)}...')

        # Intersection of the set of the names of all methods bound to this
        # class with the set of the names of any methods required by one or more
        # "collections.abc" ABCs reachable from the current FSM node, defined as
        # a frozen set to enable this intersection to be looked up as a hashable
        # key of the "node_curr.nodes_next" dictionary below.
        cls_method_collection_abc_names = frozenset(
            cls_method_names & node_curr.nodes_next_method_names)

        # If this intersection is empty, no further FSM transitions can be made.
        # In this case, the current "collections.abc" ABC validating this class
        # is the narrowest such ABC. Halt searching and return this ABC.
        if not cls_method_collection_abc_names:
            break
        # Else, this intersection is non-empty. In this case, at least one FSM
        # transition can be made. In this case, the current "collections.abc"
        # ABC validating this class is *NOT* the narrowest such ABC. Continue
        # searching for the narrowest such ABC.

        # FSM node to which these class methods transition if these class
        # methods trivially transition to an FSM node *OR* "None" otherwise.
        node_next = node_curr.nodes_next.get(cls_method_collection_abc_names)
        # print(f'FSM node {repr(node_curr.hint_factory)} intersection: {repr(node_next)}')

        # If these class methods do *NOT* trivially transition to an FSM node,
        # these class methods could still non-trivially transition to an FSM
        # node. How? By defining enough class methods to simultaneously satisfy
        # two or more FSM nodes. Consider an edge-case class satisfying both the
        # "collections.abc.Sequence" *AND* "collections.abc.Mapping" protocols,
        # for example. The "cls_method_collection_abc_names" frozen set would
        # then be the union of the keys mapping to both the
        # "collections.abc.Sequence" *AND* "collections.abc.Mapping" FSM nodes;
        # since that frozen set would be substantially larger than either of
        # those keys, that set would *NOT* be equal to either of those keys.
        # Sadly, manual iteration is the only means to detect this edge case.
        if not node_next:
            # For the frozen set of all requisite methods *AND* FSM nodes to
            # which those methods transition...
            for node_next_method_names, node_next_possible in (
                node_curr.nodes_next.items()):
                # print(f'Considering transitioning to FSM node {repr(node_next_possible.hint_factory)}...')

                # print(f'Set intersection: {node_next_method_names & cls_method_names}...')
                # print(f'Expected set: {node_next_method_names}...')

                # If the intersection of the set of the names of all methods
                # bound to this class with this frozen set of the names of all
                # methods required by this protocol is exactly this frozen set,
                # then this class defines *ALL* methods in this set and thus
                # satisfies this target "collections.abc" ABC. In this case...
                if (
                    len(node_next_method_names & cls_method_names) ==
                    len(node_next_method_names)
                ):
                    # Set the next FSM node to this node.
                    node_next = node_next_possible
                    # print(f'Transitioning to FSM node {repr(node_next.hint_factory)}...')

                    # Halt this inner iteration.
                    break

            # If these class methods still do *NOT* non-trivially transition to
            # an FSM node, halt searching immediately.
            if not node_next:
                break
            # Else, these class methods non-trivially transition to an FSM node.
            # In this case, immediately transition to this FSM node.
        # Else, these class methods trivially transition to an FSM node. In this
        # case, immediately transition to this FSM node.

        # Transition from the current to the next FSM node.
        node_curr = node_next

        # Nullify the next FSM node for safety.
        node_next = None

    # Return the narrowest "collections.abc" ABC validating this class if any
    # *OR* "None".
    return node_curr.hint_factory

# ....................{ PRIVATE ~ hints                    }....................
_FiniteStateMachine = Dict[FrozenSetStrs, '_FiniteStateMachineNode']
'''
PEP-compliant type hint matching a :mod:`collections.abc` **fine state machine
(FSM)** as a dictionary whose:

* Keys describe the **transitions** (i.e., directed edges transitioning from one
  node to another) of this FSM. Each key is a frozen set containing the names of
  two or more **dunder methods** (e.g., ``frozenset({"__iter__", "__len__"})``).
  If the object passed to the :func:`.infer_hint_type_collections_abc` function
  defines *all* of these methods, then this FSM transitions to the **node**
  (i.e., :class:`._FiniteStateMachineNode` instance) given by the value
  associated with this key in this dictionary.
* Values describe the **nodes** (i.e., :class:`._FiniteStateMachineNode`
  instances) of this FSM to follow these transitions to.
'''

# ....................{ PRIVATE ~ classes                  }....................
class _FiniteStateMachineNode(object):
    '''
    :mod:`collections.abc` **finite state machine (FSM) node** (i.e., node of a
    FSM whose transitions infer the narrowest :mod:`collections.abc` ABC
    applicable to the class passed to the
    :func:`.infer_hint_type_collections_abc` function).

    Parameters
    ----------
    hint_factory : object
        Either:

        * If this is the **start node** of this FSM, :data:`None`. The start
          node is *only* a convenient abstraction defining the initial
          transitions to actual nodes containing meaningful content.
        * Else, the type hint factory validating the passed object (e.g.,
          :class:`collections.abc.Container` if that object is a container).
    nodes_next_method_names : FrozenSet[str]
        Frozen set of the names of all **transition methods** (i.e., one or more
        methods required to be defined by the passed object for this FSM to
        transition from this node to another node). This frozen set is a
        non-negligible optimization enabling this FSM to exhibit non-amortized
        best-case constant :math:`O(1)` time complexity. Why? Because, in the
        best case, the :func:`.infer_hint_type_collections_abc` function is
        passed an object whose type defines *no* methods in this frozen set and
        thus immediately reduces to a noop.
    nodes_next : _FiniteStateMachine
        Either:

        * If this is a **stop node** of this FSM, the empty dictionary.
        * Else, this is a **transitionary node** of this FSM. In this case, this
          is the **transitionary FSM** (i.e., proper subset of the full FSM,
          mapping the transitions directed out of this node to the neighbouring
          nodes those transitions transition to) of this node.
    '''

    # ..................{ CLASS VARIABLES                    }..................
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # CAUTION: Synchronize this slots list with the implementations of:
    # * The __new__() dunder method.
    # CAUTION: Subclasses declaring uniquely subclass-specific instance
    # variables *MUST* additionally slot those variables. Subclasses violating
    # this constraint will be usable but unslotted, which defeats our purposes.
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    # Slot all instance variables defined on this object to minimize the time
    # complexity of both reading and writing variables across frequently called
    # cache dunder methods. Slotting has been shown to reduce read and write
    # costs by approximately ~10%, which is non-trivial.
    __slots__ = (
        'hint_factory',
        'nodes_next',
        'nodes_next_method_names',
    )

    # Squelch false negatives from mypy. This is absurd. This is mypy. See:
    #     https://github.com/python/mypy/issues/5941
    if TYPE_CHECKING:
        hint_factory: object
        nodes_next_method_names: FrozenSetStrs
        nodes_next: _FiniteStateMachine

    # ..................{ INITIALIZERS                       }..................
    def __init__(
        self,

        # Optional parameters.
        hint_factory: object = None,
        nodes_next: _FiniteStateMachine = FROZENDICT_EMPTY,
    ) -> None:
        '''

        Parameters
        ----------
        hint_factory: object
            Either:

            * If this is the **start node** of this FSM, :data:`None`. The start
              node is *only* a convenient abstraction defining the initial
              transitions to actual nodes containing meaningful content.
            * Else, the type hint factory validating the passed object (e.g.,
              :class:`collections.abc.Container` if that object is a container).

            Defaults to :data:`None`.
        finite_state_machine: _FiniteStateMachine
            Either:

            * If this is a **stop node** of this FSM, the empty dictionary.
            * Else, the proper subset of the full FSM mapping the transitions
              directed out of this node to the neighbouring nodes those
              transitions transition to.

            Defaults to the empty dictionary.
        '''
        assert isinstance(nodes_next, dict), (
            f'{repr(nodes_next)} not dictionary.')

        # Classify all passed parameters.
        self.hint_factory = hint_factory
        self.nodes_next = nodes_next

        # Frozen set of the names of all transition methods, initialized to the
        # empty set.
        nodes_next_method_names: Set[str] = set()

        # For the frozen set of all method names *AND* FSM nodes to which those
        # methods transition...
        for node_next_method_names, node_next in nodes_next.items():
            # Validate the types of this key-value pair.
            assert isinstance(
                node_next_method_names, frozenset), (
                f'{repr(node_next_method_names)} not frozen set.')
            assert isinstance(node_next, _FiniteStateMachineNode), (
                f'{repr(node_next)} not "_FiniteStateMachineNode".')

            # Update this set with this frozen set of all method names.
            nodes_next_method_names.update(node_next_method_names)

        # Classify this set as a frozen set for safety.
        self.nodes_next_method_names = frozenset(nodes_next_method_names)

    # ..................{ DUNDERS                            }..................
    def __repr__(self) -> str:
        '''
        Machine-readable representation of this finite state machine (FSM) node.
        '''

        return '\n'.join((
            f'{self.__class__.__name__}(',
            f'    hint_factory={repr(self.hint_factory)},',
            f'    nodes_next_method_names={repr(self.nodes_next_method_names)},',
            f'    nodes_next={repr(self.nodes_next)},',
            f')',
        ))

# ....................{ PRIVATE ~ getters                  }....................
@callable_cached
def get_finite_state_machine() -> _FiniteStateMachineNode:
    '''
    :mod:`collections.abc` **finite state machine (FSM)**.

    This getter is memoized for efficiency. Indeed, the pure-Python data
    structure created and returned by this getter is intentionally isolated to
    the body of this getter rather than defined as a global variable due to the
    cost of creating this data structure. The former approach defers that
    to the first call to the :func:`.infer_hint_factory_collections_abc`
    function, whereas the latter approach would impose that cost on the first
    importation of *anything* from the popular :mod:`beartype.bite` subpackage.
    Needless to say, the former approach is preferable.

    See Also
    --------
    :data:`._FiniteStateMachine`
        Further details.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer function-specific imports.
    from beartype.typing import (
        AsyncGenerator,
        AsyncIterable,
        AsyncIterator,
        Awaitable,
        Collection,
        Container,
        Coroutine,
        Generator,
        Iterable,
        Iterator,
        Mapping,
        MutableMapping,
        MutableSequence,
        MutableSet,
        Reversible,
        Sequence,
        Sized,
    )
    from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_12

    # ....................{ LOCALS                         }....................
    # Dictionary mapping the transitions directed out of the start node of this
    # finite state machine (FSM) to the neighbouring nodes of those transitions.
    # Note that:
    # * A "collections.abc.ByteString" FSM is intentionally *NOT* defined. Why?
    #   Because the "collections.abc.ByteString" protocol has been deprecated as
    #   of Python >= 3.12 and scheduled for removal under Python 3.14.
    # * A "collections.abc.Callable" FSM is intentionally *NOT* defined. Why?
    #   Because the sibling "_infercallable" submodule already deeply infers
    #   callable type hints for arbitrary objects satisfying the
    #   "collections.abc.Callable" protocol in a *MUCH* more intelligent manner.
    # * A "collections.abc.Hashable" FSM is intentionally *NOT* defined. Why?
    #   Because the root "object" superclass defines the __hash__() method to
    #   return the integer uniquely identifying the current object. Almost all
    #   classes thus trivially satisfy the "collections.abc.Hashable" protocol.
    #   The only exception are mutable classes redefining the "__hash__"
    #   attribute as class variables set to "None". Since the majority of
    #   classes still trivially satisfy the "collections.abc.Hashable" protocol,
    #   attempting to infer hashability (i.e., compliance with this protocol)
    #   would uselessly infer the type hint factory for almost all classes as
    #   the universally broad and unsubscriptable "collections.abc.Hashable" ABC
    #   rather than as those specifically narrow classes themselves. Since
    #   everyone prefers and expects specifically narrow type hints rather than
    #   universally broad and unsubscriptable type hints, we ignore hashability.
    # * Dictionary-based views FSMs are intentionally *NOT* defined. Why?
    #   Because:
    #   * The "collections.abc.MappingView" protocol is indistinguishable from
    #     the "collections.abc.Sized" protocol.
    #   * The "collections.abc.ItemsView" protocol are indistinguishable from
    #     the "collections.abc.Collection" protocol.
    #   * The "collections.abc.KeysView" and "ValuesView" protocols are
    #     indistinguishable from the "collections.abc.Set" protocol.
    _START_NODE_NODES_NEXT: _FiniteStateMachine = {
        # ....................{ CONTAINER                  }....................
        # "collections.abc.Container" FSM, intentionally defined first to
        # prioritize this most useful ABC over relatively useless alternatives.
        frozenset(('__contains__',)): _FiniteStateMachineNode(
            hint_factory=Container,
            nodes_next={
                # ....................{ COLLECTION         }....................
                # "collections.abc.Collection" FSM.
                frozenset(('__iter__', '__len__',)): _FiniteStateMachineNode(
                    hint_factory=Collection,
                    nodes_next={
                        # ....................{ SEQUENCE   }....................
                        # "collections.abc.Sequence" FSM, intentionally defined
                        # first to prioritize this most useful ABC over
                        # relatively useless alternatives.
                        frozenset((
                            '__getitem__',
                            '__reversed__',
                            'count',
                            'index',
                        )): _FiniteStateMachineNode(
                            hint_factory=Sequence,
                            nodes_next={
                                # "collections.abc.MutableSequence" FSM.
                                frozenset((
                                    '__delitem__',
                                    '__iadd__',
                                    '__setitem__',
                                    'append',
                                    'clear',
                                    'extend',
                                    'insert',
                                    'pop',
                                    'remove',
                                    'reverse',
                                )): _FiniteStateMachineNode(
                                    hint_factory=MutableSequence),
                            },
                        ),
                        # ....................{ MAPPING    }....................
                        # "collections.abc.Mapping" FSM, intentionally defined
                        # nearly first to prioritize this more useful ABC over
                        # relatively useless alternatives.
                        frozenset((
                            '__eq__',
                            '__ne__',
                            '__getitem__',
                            'get',
                            'items',
                            'keys',
                            'values',
                        )): _FiniteStateMachineNode(
                            hint_factory=Mapping,
                            nodes_next={
                                # "collections.abc.MutableMapping" FSM.
                                frozenset((
                                    '__delitem__',
                                    '__setitem__',
                                    'clear',
                                    'pop',
                                    'popitem',
                                    'setdefault',
                                    'update',
                                )): _FiniteStateMachineNode(
                                    hint_factory=MutableMapping),
                            },
                        ),
                        # ....................{ SET        }....................
                        # "collections.abc.Set" FSM, intentionally defined
                        # midway to prioritize this more useful ABC over
                        # relatively useless alternatives.
                        frozenset((
                            '__and__',
                            '__eq__',
                            '__ge__',
                            '__gt__',
                            '__le__',
                            '__lt__',
                            '__ne__',
                            '__or__',
                            '__sub__',
                            '__xor__',
                            'isdisjoint',
                        )): _FiniteStateMachineNode(
                            hint_factory=Set,
                            nodes_next={
                                # "collections.abc.MutableSet" FSM.
                                frozenset((
                                    '__iand__',
                                    '__ior__',
                                    '__isub__',
                                    '__ixor__',
                                    'clear',
                                    'pop',
                                    'remove',
                                )): _FiniteStateMachineNode(
                                    hint_factory=MutableSet),
                            },
                        ),
                    },
                ),
            },
        ),
        # ....................{ ITERABLE                   }....................
        # "collections.abc.Iterable" FSM, intentionally defined nearly first to
        # prioritize this more useful ABC over relatively useless alternatives.
        frozenset(('__iter__',)): _FiniteStateMachineNode(
            hint_factory=Iterable,
            nodes_next={
                # "collections.abc.Iterator" FSM, intentionally defined first to
                # prioritize this more useful ABC over relatively useless
                # alternatives.
                frozenset(('__next__',)): _FiniteStateMachineNode(
                    hint_factory=Iterator,
                    nodes_next={
                        # "collections.abc.Generator" FSM.
                        frozenset((
                            'close',
                            'send',
                            'throw',
                        )): _FiniteStateMachineNode(
                            hint_factory=Generator,
                            nodes_next={
                            },
                        ),
                    },
                ),
                # ....................{ REVERSIBLE         }....................
                # "collections.abc.Reversible" FSM, intentionally defined nearly
                # last to deprioritize this relatively useless ABC over more
                # useful alternatives.
                frozenset(('__reversed__',)): _FiniteStateMachineNode(
                    hint_factory=Reversible),
            },
        ),
        # ....................{ AWAITABLE                  }....................
        # "collections.abc.Awaitable" FSM, intentionally defined midway to
        # prioritize this more useful ABC over relatively useless alternatives.
        frozenset(('__await__',)): _FiniteStateMachineNode(
            hint_factory=Awaitable,
            nodes_next={
                # "collections.abc.Coroutine" FSM.
                frozenset((
                    'close',
                    'send',
                    'throw',
                )): _FiniteStateMachineNode(hint_factory=Coroutine),
            },
        ),
        # ....................{ ASYNCITERABLE              }....................
        # "collections.abc.AsyncIterable" FSM, intentionally defined nearly last
        # to prioritize this less useful ABC over relatively useless
        # alternatives.
        frozenset(('__aiter__',)): _FiniteStateMachineNode(
            hint_factory=AsyncIterable,
            nodes_next={
                # "collections.abc.AsyncIterator" FSM.
                frozenset(('__anext__',)): _FiniteStateMachineNode(
                    hint_factory=AsyncIterator,
                    nodes_next={
                        # "collections.abc.AsyncGenerator" FSM.
                        frozenset((
                            'aclose',
                            'asend',
                            'athrow',
                        )): _FiniteStateMachineNode(
                            hint_factory=AsyncGenerator),
                    },
                ),
            },
        ),
        # ....................{ SIZED                      }....................
        # "collections.abc.Sized" FSM, intentionally defined last to
        # deprioritize this almost entirely useless ABC over more useful
        # alternatives. Since *ALMOST* all classes satisfying this ABC also
        # satisfy the "collections.abc.Collection" ABC, this ABC is (in and of
        # itself) both largely useless and exceedingly uncommon.
        frozenset(('__len__',)): _FiniteStateMachineNode(hint_factory=Sized),
    }

    # If the active Python interpreter targets Python >= 3.12...
    if IS_PYTHON_AT_LEAST_3_12:
        # Defer version-specific imports.
        from collections.abc import Buffer  # type: ignore[attr-defined]

        # Add the "collections.abc.Buffer" FSM.
        _START_NODE_NODES_NEXT[frozenset(('__buffer__',))] = (
            _FiniteStateMachineNode(hint_factory=Buffer))
    # Else, the active Python interpreter targets Python < 3.12.

    # "collections.abc" finite state machine (FSM).
    START_NODE = _FiniteStateMachineNode(nodes_next=_START_NODE_NODES_NEXT)

    # Return this FSM.
    return START_NODE



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Test-wide **type hint container data submodule.**

This submodule predefines low-level global constants exercising known edge
cases on behalf of higher-level unit test submodules -- including PEP-compliant
type hints, PEP-noncompliant type hint, and objects satisfying neither.
'''

# ....................{ TODO                               }....................
#FIXME: This submodule is largely vestigial. As time permits (so, never),
#refactor *ALL* remaining globals #below into either:
#* Comparable fixtures in the sibling "data_hintfixture" submodule.
#* Comparable globals in the parent "data_type" submodule.
#
#There's *A LOT* of weird stuff going on below. So it goes, QA friends. *sigh*

# ....................{ IMPORTS                            }....................
from beartype._cave._cavefast import (
    AnyType,
    NoneType,
)
from beartype._cave._cavemap import NoneTypeOr

# ....................{ NON-HINTS ~ sets                   }....................
NOT_HINTS_HASHABLE = frozenset((
    # Scalar that is neither a type nor string (i.e., forward reference).
    0.12345678910111213141516,
    # Empty tuple.
    (),
    # Tuple containing a scalar that is neither a type nor string.
    (list, 'list', 0xFEEDFACE, NoneType,),
))
'''
Frozen set of various objects that are hashable but nonetheless unsupported by
the :func:`beartype.beartype` decorator as valid type hints.
'''

# ....................{ NON-HINTS ~ tuples                 }....................
NOT_HINTS_UNHASHABLE = (
    # Dictionary.
    {'For all things turn to barrenness':
     'In the dim glass the demons hold,',},
    # List.
    ['The glass of outer weariness,',
     'Made when God slept in times of old.',],
    # Set.
    {'There, through the broken branches, go',
     'The ravens of unresting thought;',},
)
'''
Tuple of various objects that are unhashable and thus unsupported by the
:func:`beartype.beartype` decorator as valid type hints.

Since sets *cannot* by design contain unhashable objects, this container is
defined as a tuple rather than a set.
'''


NOT_HINTS = tuple(NOT_HINTS_HASHABLE) + NOT_HINTS_UNHASHABLE
'''
Tuple of various objects unsupported by the :func:`beartype.beartype` decorator
as valid type hints, including both hashable and unhashable unsupported
objects.

Since sets *cannot* by design contain unhashable objects, this container is
defined as a tuple rather than a set.
'''

# ....................{ NON-PEP ~ classes                  }....................
class NonpepCustom(str):
    '''
    PEP-noncompliant user-defined class subclassing an arbitrary superclass.
    '''

    pass


class NonpepCustomFakeTyping(object):
    '''
    PEP-noncompliant user-defined class subclassing an arbitrary superclass
    erroneously masquerading as a :mod:`typing` class.

    Specifically, this class:

    * Defines the :meth:`__repr__` dunder method to return a string prefixed by
      the substring ``"typing."``.
    * Defines the :attr:`__module__` dunder attribute to be the fully-qualified
      module name ``"typing"``.
    '''

    def __repr__(self) -> str:
        return 'typing.FakeTypingType'
NonpepCustomFakeTyping.__module__ = 'typing'

# ....................{ NON-PEP ~ sets                     }....................
# Note that we intentionally omit the "NonpepCustomFakeTyping" class here, as
# that class masquerades too well as a "typing" class -- so well, in fact, that
# "beartype._util.hint.pep.utilpeptest" functions are incapable of reasonably
# distinguishing instances of that class from actual "typing" type hints.
HINTS_NONPEP = frozenset((
    # Builtin container type.
    list,
    # Builtin scalar type.
    str,
    # User-defined type.
    NonpepCustom,
    # Non-empty tuple containing two types.
    NoneTypeOr[AnyType],
    # Non-empty tuple containing two types and a fully-qualified forward
    # reference.
    (int, 'beartype._cave._cavefast.NoneType', set)
))
'''
Frozen set of PEP-noncompliant type hints exercising well-known edge cases.
'''


HINTS_NONPEP_UNIGNORABLE = (
    # PEP-noncompliant type.
    str,
    # PEP-noncompliant tuple of types.
    NoneTypeOr[AnyType],
)
'''
Frozen set of **unignorable PEP-noncompliant type hints** (i.e.,
PEP-noncompliant type hints that are *not* ignorable).
'''

# ....................{ NOT ~ sets                         }....................
NOT_HINTS_PEP = (
    # PEP-noncompliant type hints.
    HINTS_NONPEP |
    # Hashable objects invalid as type hints (e.g., scalars).
    NOT_HINTS_HASHABLE
)
'''
Tuple of various objects that are *not* PEP-compliant type hints exercising
well-known edge cases.
'''
